# Variables and functions to help with plotting functions
# These do not do the plotting themselves
from __future__ import annotations

import glob
import logging
import os
from typing import TYPE_CHECKING

import numpy as np

from proteus.utils.helper import mol_to_ele

log = logging.getLogger("fwl."+__name__)

if TYPE_CHECKING:
    from proteus import Proteus

# Standard plotting colours
_preset_colours  = {
    # Default colour
    "_fallback": "#ff00ff",

    # Volatile gases
    "H2O": "#027FB1",
    "CO2": "#D24901",
    "H2" : "#008C01",
    "CH4": "#C720DD",
    "CO" : "#D1AC02",
    "N2" : "#870036",
    "S2" : "#FF8FA1",
    "SO2": "#00008B",

    # Volatile elements
    "H": "#0000aa",
    "C": "#ff0000",
    "O": "#00dd00",
    "N": "#ffaa00",
    "S": "#ff22ff",
    "P": "#33ccff",

    # refractory elements
    "Fe": "#888888",
    "Si": "#aa2277",
    "Mg": "#996633",

    # Energy fluxes
    "OLR":   "#dc143c",
    "ASF":   "#4169e1",
    "sct":   "#2e8b57",
    "tidal": "#daa520",
    "radio": "#1e90ff",

    # Model components
    "atm"     : "#444444",
    "int"     : "#ff7f0e",
    "core"    : "#aa3077",
    "atm_bkg" : "#f2faff",
    "int_bkg" : "#fffaf2",
    "cor_bkg" : "#efefef",
}


def _generate_colour(gas:str):
    """
    Systematically generate a colour for a gas, from its composition.

    This method for mixing colours was taken from AGNI's phys.jl
    """

    # Break into atoms
    try:
        atoms = mol_to_ele(gas)
    except ValueError:
        log.warning(f"Using fallback colour for '{gas}'")
        return _preset_colours["_fallback"]

    # Red, green, blue components
    red = 0.0
    gre = 0.0
    blu = 0.0

    # For each atom, add the contriution of its rgb components
    for e in atoms.keys():
        red += int(_preset_colours[e][1:2],base=16)*atoms[e]
        gre += int(_preset_colours[e][3:4],base=16)*atoms[e]
        blu += int(_preset_colours[e][5:6],base=16)*atoms[e]

    # Normalisation constant
    norm = max((red,gre,blu))

    # Normalise colours to 0-255
    red = int(255*red/norm)
    gre = int(255*gre/norm)
    blu = int(255*blu/norm)

    # Prevent the colour getting too close to white, which is hard to see on a plot
    c_lim = 220
    if red+gre+blu > c_lim*3:
        scale = c_lim/255.0
        red = int(scale*red)
        gre = int(scale*gre)
        blu = int(scale*blu)

    # Convert to hex string
    colour = f"#{red:02x}{gre:02x}{blu:02x}"

    return colour


def get_colour(thing:str):
    """
    Get a colour for something which needs one (e.g. for plotting a particular gas)
    """

    # Try getting it from dictionary
    try:
        colour = _preset_colours[thing]

    # Otherwise, generate it systematically
    except KeyError:
        colour = _generate_colour(thing)

    return colour

def latexify(gas:str):
    """
    Convert gas name to latex-formatted string
    """

    out = ""
    for c in gas:
        c = str(c)
        if c.isnumeric():
            out += r"$_%s$"%c
        else:
            out += c
    return out

# Bandpasses for instrumentation of interest [units of microns, um]
observer_bands = {

    # https://jwst-docs.stsci.edu/jwst-mid-infrared-instrument/miri-instrumentation/miri-filters-and-dispersers
    "MIRI" : {
        "F560W":   [5.054, 6.171],
        "F770W":   [6.581, 8.687],
        "F1000W":  [9.023, 10.891],
        "F1130W":  [10.953, 11.667],
        "F1280W":  [11.588, 14.115],
        "F1500W":  [13.527, 16.64],
        "F1800W":  [16.519, 19.502],
        "F2100W":  [18.477, 23.159],
        "F2550W":  [23.301, 26.733],
        "LRS":     [5, 14],
    },

    # https://jwst-docs.stsci.edu/jwst-near-infrared-spectrograph/nirspec-instrumentation/nirspec-dispersers-and-filters#
    "NIRSpec" : {
        "F070LP": [0.70,1.27],
        "F100LP": [0.97,1.84],
        "F170LP": [1.66,3.07],
        "F290LP": [2.87,5.10],
        "PRISM" : [0.60,5.30],
    },

    # https://jwst-docs.stsci.edu/jwst-near-infrared-imager-and-slitless-spectrograph#gsc.tab=0
    "NIRISS" : {
        "SOSS": [0.6, 2.8],
        "WFSS": [0.8, 2.2],
        "AMI":  [2.8, 4.8]
    },

    # https://jwst-docs.stsci.edu/jwst-near-infrared-camera/nircam-instrumentation/nircam-filters
    "NIRCam" : {
        "Short": [0.6, 2.3],
        "Long":  [2.4, 5.0],
    },

    # https://www.esa.int/Science_Exploration/Space_Science/Ariel/Ariel_s_instruments
    "ARIEL" : {
        "AIRS0": [1.95, 3.9],
        "AIRS1": [3.9, 7.8]
    },

    # https://en.wikipedia.org/wiki/Infrared_astronomy
    # https://www.eso.org/sci/facilities/paranal/instruments/crires/inst.html
    # https://www.eso.org/sci/facilities/paranal/instruments/gravity/overview.html
    "IR" : {
        "R": [0.65 , 1.0 ],
        "J": [1.115, 1.362 ],
        "H": [1.423, 1.769],
        "K": [1.972, 2.624],
        "L": [2.869, 4.188],
        "M": [4.6  , 5.0 ],
        "N": [7.5  , 14.5],
        "Q": [17.0 , 25.0],
        "Z": [28.0 , 40.0],
    },

    # https://link.springer.com/article/10.1007/s10686-020-09660-1
    "PLATO" : {
        "blue": [0.500, 0.675],
        "red" : [0.675, 1.125]
    },

    # https://doi.org/10.1051/0004-6361/202140366
    "LIFE" : {
        "LIFE": [4.0, 18.5]
    },

    # https://ntrs.nasa.gov/api/citations/20240006497/downloads/HWO%20Engineering%20View%20Status%20Plans%20Opportunities.pdf
    "HWO" : {
        "Coronograph":      [0.4, 1.8],
        "Highres imager":   [0.2, 2.5],
        "Spectrograph":     [0.1, 1.0]
    },

    # https://www.gemini.edu/instrumentation/maroon-x
    # https://www.gemini.edu/instrumentation/igrins-2
    "GEMINI-N" : {
        "MAROON-X": [0.5, 0.92],
        "IGRINS-2": [1.49, 2.46]
    },

    # https://www.eso.org/sci/facilities/paranal/instruments/sphere.html
    # https://www.eso.org/sci/facilities/paranal/instruments/espresso/overview.html
    "VLT" : {
        "ESPRESSO": [0.38, 0.788],
        "SPHERE": [0.95, 2.32]
    },

    # https://carmenes.caha.es/ext/instrument/index.html
    "CARMENES": {
        "CARMENES": [0.520, 1.710],
    },

    # https://www.tng.iac.es/instruments/harps/
    "HARPS" : {
        "HARPS-N":   [0.383, 0.690],
    },

    # https://noirlab.edu/public/programs/kitt-peak-national-observatory/wiyn-35m-telescope/neid/
    "NEID" : {
        "NEID": [0.38, 0.93],
    },

    # https://elt.eso.org/instrument/
    "ELT" : {
        "HARMONI": [0.47, 2.45],
        "MICADO":  [0.8,  2.4],
        "ANDES":   [0.40, 1.80]
    }

}

# https://stackoverflow.com/questions/13490292/format-number-using-latex-notation-in-python
def latex_float(f):
    float_str = "{0:.2g}".format(f)
    if "e" in float_str:
        base, exponent = float_str.split("e")
        return r"${0} \times 10^{{{1}}}$".format(base, int(exponent))
    else:
        return float_str

def sample_times(times:list, nsamp:int, tmin:float=1.0):
    from proteus.utils.helper import find_nearest

    # check count
    if len(times) <= nsamp:
        out_t, out_i = np.unique(times, return_index=True)
        return list(out_t), list(out_i)

    # lower limit
    tmin = max(tmin,np.amin(times))
    tmin = min(tmin, np.amax(times))
    tmin = max(tmin, 1.0)
    # upper limit
    tmax = max(tmin+1, np.amax(times))

    # do not allow times outside range
    allowed_times = [int(x) for x in times if tmin<=x<=tmax]

    # get samples on log-time scale
    sample_t = []
    sample_i = []
    for s in np.logspace(np.log10(tmin),np.log10(tmax),nsamp): # Sample on log-scale

        remaining = [int(v) for v in set(allowed_times) - set(sample_t)]
        if len(remaining) == 0:
            break

        # Get next nearest time
        val,_ = find_nearest(remaining,s)
        sample_t.append(int(val))

        # Get the index of this time in the original array
        _,idx = find_nearest(times,val)
        sample_i.append(int(idx))

    # sort output
    mask = np.argsort(sample_t)
    out_t, out_i = [], []
    for i in mask:
        out_t.append(sample_t[i])
        out_i.append(sample_i[i])

    return out_t, out_i


def sample_output(handler: Proteus, extension:str = ".nc", tmin:float = 1.0, nsamp:int=8):

    # get all files
    files = glob.glob(os.path.join(handler.directories["output"], "data", "*"+extension))
    if len(files) < 1:
        return []

    # get times
    times = [int(f.split("/")[-1].split(extension)[0]) for f in files]

    out_t, out_i = sample_times(times, nsamp, tmin=tmin)
    out_f = [files[i] for i in out_i]

    # return times and file paths
    return out_t, out_f

class MyFuncFormatter( object ):

    '''the default function formatter from
       matplotlib.ticker.FuncFormatter(func) only accepts two
       arguments, which is not enough to scale an arcsinh function.
       But by creating our own class here we can attach the scaling
       to the object which can then be accessed in __call__'''

    def __init__( self, arcsinh_scale ):
        self.const = arcsinh_scale

    def ascale( self, xx ):
        '''map input to log-like values (scaled arcsinh)'''
        yy = np.arcsinh( xx*self.const )
        return yy

    def _invascale( self, yy ):
        '''map input from log-like values (inverse transform)'''
        xx = np.sinh(yy) / self.const
        return xx

    def _sci_notation( self, num, decimal_digits=1, precision=None, exponent=None):
        """
        Returns a string representation of the scientific
        notation of the given number formatted for use with
        LaTeX or Mathtext, with specified number of significant
        decimal digits and precision (number of decimal digits
        to show). The exponent to be used can also be specified
        explicitly.
        """

        # plotting zero is useful to emphasize that we are plotting both
        # positive and negative values, e.g. for the heat fluxes
        if num==0:
            fmt = r"$0$"
            return fmt

        if not exponent:
            exponent = abs(num)
            exponent = np.log10( exponent )
            exponent = np.floor( exponent )
            exponent = int( exponent )

        coeff = round(num / float(10**exponent), decimal_digits)
        # sometimes, probably due to floating point precision? the coeff
        # is not less than ten.  Correct for that here
        if np.abs(coeff) >= 10.0:
            coeff /= 10.0
            exponent += 1
        if not precision:
            precision = decimal_digits

        if coeff < 0.0:
            fmt = r"$-10^{{{0}}}$".format(exponent)
            #fmt= r"${{{0}}}$".format(exponent)
        else:
            fmt = r"$10^{{{0}}}$".format(exponent)

        return fmt
        #return r"${0:.{2}f}\cdot10^{{{1:d}}}$".format(coeff, exponent, precision)

    def __call__( self, x, pos ):
        y = self._invascale( x )
        fmt = self._sci_notation( y, 0 )
        return fmt



# Generic stellar evolution wrapper
from __future__ import annotations

import logging
import os
import shutil
from typing import TYPE_CHECKING

import numpy as np

from proteus.utils.constants import AU, M_sun, R_sun, const_sigma

log = logging.getLogger("fwl."+__name__)

if TYPE_CHECKING:
    from proteus import Proteus
    from proteus.config import Config

def init_star(handler:Proteus):
    '''
    Star-related things to be done when the simulation begins.
    '''

    log.info("Preparing stellar model")

    # Dummy star module does not require preparation

    # Prepare MORS
    if handler.config.star.module == 'mors':
        import mors

        # Path to the modern spectrum
        #   i.e. that observed by a telescope, or derived from various observations.
        #   This is what we download from OSF.
        star_modern_path = os.path.join(handler.directories["fwl"],
                                        handler.config.star.mors.spec)

        # Copy modern spectrum to output folder, for posterity.
        star_backup_path = os.path.join(handler.directories["output"], "-1.sflux")
        shutil.copyfile(star_modern_path, star_backup_path)

        match handler.config.star.mors.tracks:

            case 'spada':
                # creates track data
                handler.stellar_track = mors.Star(Mstar = handler.config.star.mass,
                                                  percentile = handler.config.star.mors.rot_pctle)

                # load modern spectrum
                # calculate band-integrated fluxes
                handler.star_struct = mors.spec.Spectrum()
                handler.star_struct.LoadTSV(star_modern_path)
                handler.star_struct.CalcBandFluxes()

                # calculate other properties from modern spectrum
                handler.star_props = get_spada_synthesis_properties(
                    handler.stellar_track,
                    handler.config.star.mors.age_now * 1000, # convert Gyr to Myr
                )

            case 'baraffe':
                # creates track data
                handler.stellar_track = mors.BaraffeTrack(handler.config.star.mass)

                # load modern spectrum
                handler.star_modern_wl, handler.star_modern_fl = mors.ModernSpectrumLoad(
                    star_modern_path, star_backup_path
                )

                # calculate other properties from modern spectrum (here bolometric luminosity only)
                handler.star_props = handler.stellar_track.BaraffeLuminosity(handler.config.star.mors.age_now * 1e9)

def get_spada_synthesis_properties(spada_track, age: float):
    """Calculate properties of star for spectrum synthesis
       Mimic the GetProperties function of mors synthesis module using whole track data.

    Parameters
    ----------
        spada_track : star object
            Star object containing track data.
        age : float
            Stellar age  [Myr]

    Returns
    ----------
        out : dict
            Dictionary of radius [m], Teff [K], and band fluxes at 1 AU [erg s-1 cm-2]
    """

    import mors.spectrum as spec

    out = {}
    out["age"] = age
    out["Rstar"] = spada_track.Value(age, "Rstar") * R_sun #[m]
    out["Teff"] = spada_track.Value(age, "Teff") #[K]

    # Luminosities (erg s-1)
    out["L_bo"] = spada_track.Value(age, "Lbol")
    out["L_xr"] = spada_track.Value(age, "Lx")
    out["L_e1"] = spada_track.Value(age, "Leuv1")
    out["L_e2"] = spada_track.Value(age, "Leuv2")

    # Fluxes at 1 AU
    area = (4.0 * np.pi * AU * AU)
    for k in ["bo","xr","e1","e2"]:
        out["F_"+k] = out["L_"+k]/area

    # Get flux from Planckian band
    wl_pl = np.logspace(np.log10(spec.bands_limits["pl"][0]), np.log10(spec.bands_limits["pl"][1]), 1000)
    fl_pl = spec.PlanckFunction_surf(wl_pl, out["Teff"])
    fl_pl = spec.ScaleTo1AU(fl_pl, out["Rstar"])
    out["F_pl"] = np.trapz(fl_pl, wl_pl)
    out["L_pl"] = out["F_pl"] * area

    # Get flux of UV band from remainder
    out["F_uv"] = out["F_bo"] - out["F_xr"] - out["F_e1"] - out["F_e2"] - out["F_pl"]
    out["L_uv"] = out["F_uv"] * area

    return out

def get_new_spectrum(t_star:float, config:Config,
                     star_struct_modern=None, star_props_modern=None,
                     stellar_track=None, modern_wl=None, modern_fl=None):
    '''
    Get new stellar spectrum at 1 AU.
    '''

    log.debug("Get new stellar spectrum (star age = %g Gyr)"%(t_star/1e9))

    # Dummy case
    if config.star.module == 'dummy':
        from proteus.star.dummy import generate_spectrum
        wl, fl = generate_spectrum(config.star.dummy.Teff, config.star.dummy.radius * R_sun)

    # Mors cases
    elif config.star.module == 'mors':

        import mors

        match config.star.mors.tracks:
            case 'spada':
                star_props_hist = get_spada_synthesis_properties(stellar_track, t_star/1e6)
                assert star_struct_modern
                assert star_props_modern
                synthetic = mors.synthesis.CalcScaledSpectrumFromProps(
                    modern_spec=star_struct_modern, modern_dict=star_props_modern, historical_dict=star_props_hist)
                fl = synthetic.fl
                wl = synthetic.wl
            case 'baraffe':
                fl = stellar_track.BaraffeSpectrumCalc(
                        t_star, star_props_modern, modern_fl)
                wl = modern_wl

    return wl, fl

def scale_spectrum_to_toa(fl_arr, sep:float):
    '''
    Scale stellar fluxes from 1 AU to top of atmosphere
    '''
    return np.array(fl_arr) * ( (AU / sep)**2 )

def write_spectrum(wl_arr, fl_arr, hf_row:dict, output_dir:str):
    '''
    Write stellar spectrum to file.
    '''

    log.debug("Writing stellar spectrum to file")

    # Header information
    header = (
        "# WL(nm)\t Flux(ergs/cm**2/s/nm)   Stellar flux at t_star = %.2e yr"
        % hf_row["age_star"]
    )

    # Write to TSV file
    np.savetxt(
        os.path.join(output_dir, "data", "%d.sflux" % hf_row["Time"]),
        np.array([wl_arr, fl_arr]).T,
        header=header,
        comments="",
        fmt="%.8e",
        delimiter="\t",
    )

def update_stellar_quantities(hf_row:dict, config:Config, stellar_track=None):

    # Update value for star's radius and mass
    log.info("Update stellar radius and mass")
    update_stellar_radius(hf_row, config, stellar_track)
    update_stellar_mass(hf_row, config)

    # Update value for instellation flux
    log.info("Update instellation")
    update_instellation(hf_row, config, stellar_track)

    # Calculate new eqm temperature
    log.info("Update equilibrium temperature")
    update_equilibrium_temperature(hf_row, config)

    # Calculate new skin temperature
    # Assuming a grey stratosphere in radiative eqm (https://doi.org/10.5194/esd-7-697-2016)
    hf_row["T_skin"] = hf_row["T_eqm"] * (0.5**0.25)

def update_stellar_mass(hf_row:dict, config:Config):
    '''
    Update stellar mass in hf_row, stored in SI units.
    '''
    hf_row["M_star"] = config.star.mass * M_sun

def update_stellar_radius(hf_row:dict, config:Config, stellar_track=None):
    '''
    Update stellar radius in hf_row, stored in SI units.
    '''

    # Dummy case
    if config.star.module == 'dummy':
        R_star = config.star.dummy.radius

    # Mors cases
    elif config.star.module == 'mors':

        # which track?
        match config.star.mors.tracks:
            case 'spada':
                R_star = stellar_track.Value(hf_row["age_star"] / 1e6, "Rstar")
            case 'baraffe':
                R_star = stellar_track.BaraffeStellarRadius(hf_row["age_star"])

    # Dimensionalise and store in dictionary
    hf_row["R_star"] = R_star * R_sun

def update_instellation(hf_row:dict, config:Config, stellar_track=None):
    '''
    Update hf_row value of bolometric stellar flux impinging upon the planet.
    '''

    # Dummy case
    if config.star.module == 'dummy':
        from proteus.star.dummy import calc_instellation
        S_0 = calc_instellation(config.star.dummy.Teff, hf_row["R_star"], hf_row["separation"])

    # Mors cases
    elif config.star.module == 'mors':

        # which track?
        match config.star.mors.tracks:
            case 'spada':
                S_0 = stellar_track.Value(hf_row["age_star"] / 1e6, "Lbol") * 1e-7 \
                        / (4.0 * np.pi * hf_row["separation"]**2.0 )

            case 'baraffe':
                S_0 = stellar_track.BaraffeSolarConstant(hf_row["age_star"],
                                                    hf_row["separation"]/AU)

    # Update hf_row dictionary
    hf_row["F_ins"] = S_0

def update_equilibrium_temperature(hf_row:dict, config:Config):
    '''
    Calculate planetary equilibrium temperature.
    '''

    # Absorbed stellar flux
    F_asf = hf_row["F_ins"] * config.orbit.s0_factor * (1-config.atmos_clim.albedo_pl)

    # Planetary equilibrium temperature
    hf_row["T_eqm"] = (F_asf / const_sigma)**(1.0/4.0)



from __future__ import annotations

from attrs import define, field
from attrs.validators import ge, gt, in_, lt


def no_radio_if_dummy(instance, attribute, value):
    if (instance.module == 'dummy') and value:
        raise ValueError("Radiogenic heating is not supported by the dummy interior module")

@define
class Interior:
    """Magma ocean model selection and parameters.

    Attributes
    ----------
    grain_size: float
        Crystal settling grain size [m].
    F_initial: float
        Initial heat flux guess [W m-2].
    radiogenic_heat: bool
        Include radiogenic heat production?
    module: str
        Module for simulating the magma ocean. Choices: 'spider', 'aragog', 'dummy'.
    spider: Spider
        Parameters for running the SPIDER module.
    aragog: Aragog
        Parameters for running the aragog module.
    dummy: Dummy
        Parameters for running the dummy module.
    """
    grain_size: float = field(validator=gt(0))
    F_initial: float = field(validator=gt(0))
    radiogenic_heat: bool = field(validator=no_radio_if_dummy)
    tidal_heat: bool

    module: str = field(validator=in_(('spider', 'aragog', 'dummy')))

    spider: Spider
    aragog: Aragog
    dummy: Dummy


@define
class Spider:
    """Parameters for SPIDER module.

    Attributes
    ----------
    num_levels: int
        Number of SPIDER grid levels.
    mixing_length: int
        Parameterisation used to determine convective mixing length.
    tolerance: float
        Solver tolerance.
    tsurf_atol: float
        Absolute tolerance on change in T_mantle during a single interior iteration.
    tsurf_rtol: float
        Relative tolerance on change in T_mantle during a single interior iteration.
    ini_entropy: float
        Initial specific surface entropy [J K-1 kg-1].
    ini_dsdr: float
        Initial interior specific entropy gradient [J K-1 kg-1 m-1].
    """
    num_levels: int  = field(validator=ge(40))
    mixing_length: int  = field(validator=in_((1,2)))
    tolerance: float = field(validator=gt(0))
    tsurf_atol: float = field(validator=gt(0))
    tsurf_rtol: float = field(validator=gt(0))
    ini_entropy: float = field(validator=gt(0))
    ini_dsdr: float = field(validator=lt(0))


@define
class Aragog:
    """Parameters for Aragog module.

    Attributes
    ----------
    num_levels: int
        Number of Aragog grid levels (basic mesh).
    tolerance: float
        Solver tolerance.
    ini_tmagma: float
        Initial magma surface temperature [K].
    """

    ini_tmagma: float = field(validator=gt(0))
    num_levels: int  = field(validator=ge(40))
    tolerance: float

@define
class Dummy:
    """Parameters for Dummy interior module.

    Attributes
    ----------
    ini_tmagma: float
        Initial magma surface temperature [K].
    """

    ini_tmagma: float = field(validator=gt(0))



from __future__ import annotations

from attrs import define, field
from attrs.validators import ge, gt, in_, lt

from ._converters import none_if_none


@define
class Orbit:
    """Planetary orbital parameters.

    Attributes
    ----------
    semimajoraxis: float
        Semi-major axis of the planet's orbit [AU].
    eccentricity: float
        Eccentricity of the planet's orbit.
    zenith_angle: float
        Characteristic angle of incoming stellar radiation, relative to the zenith [deg].
    s0_factor: float
        Scale factor applies to incoming stellar radiation to represent planetary rotation and heat redistribution.
    module: str | None
        Select orbit module to use. Choices: 'none', 'dummy'.
    """
    semimajoraxis: float = field(validator=gt(0))
    eccentricity: float = field(validator=(
        ge(0),
        lt(1),
    ))
    zenith_angle: float  = field(validator=(
        ge(0),
        lt(90),
    ))
    s0_factor: float = field(validator=gt(0))

    module: str | None = field(
        validator=in_((None, 'dummy')),
        converter=none_if_none,
    )

    dummy: OrbitDummy


@define
class OrbitDummy:
    """Dummy orbit module.

    Attributes
    ----------
    H_tide: float
        Fixed global heating rate from tides [W kg-1].
    """
    H_tide: float = field(validator=ge(0))



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from cmcrameri import cm
from mpl_toolkits.axes_grid1 import make_axes_locatable

from proteus.interior.wrapper import read_interior_data
from proteus.utils.plot import sample_output

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)

def plot_interior_cmesh(output_dir: str, times: list | np.ndarray, data: list,
                        module: str, use_contour: bool=True,
                        cblevels: int=24, numticks: int=5, plot_format: str="pdf"):

    if len(data) < 3:
        log.warning("Too few samples to make interior_cmesh plot")
        return

    log.info("Plot interior colourmesh")

    # Downsample data?
    if len(data) < 1000:
        stride = int(1)
    else:
        stride = int(15)
    data = data[::stride]
    times = times[::stride]
    nfiles = len(data)

    # Initialise plot
    scale = 1.0
    fig,axs = plt.subplots(4,1, sharex=True, figsize=(5*scale,7*scale))
    ax1,ax2,ax3,ax4 = axs
    for ax in axs:
        ax.set_ylabel("Pressure [GPa]")
    ax4.set_xlabel("Time [yr]")

    # Pressure [GPa] grid
    ds = data[0]
    if module == "aragog":
        arr_yb = ds["pres_b"]
    elif module == "spider":
        arr_yb = ds.get_dict_values(['data','pressure_b']) * 1e-9
    nlev_b   = len(arr_yb)

    # Create 2D arrays
    arr_z1 = np.zeros((nfiles,nlev_b),dtype=float)
    arr_z2 = np.zeros((nfiles,nlev_b),dtype=float)
    arr_z3 = np.zeros((nfiles,nlev_b),dtype=float)
    arr_z4 = np.zeros((nfiles,nlev_b),dtype=float)

    # Populate 2D arrays
    for i in range(nfiles):

        # Get 1D data for this index
        ds = data[i]
        if module == "aragog":
            y_tmp = ds["temp_b"][:]
            y_phi = ds["phi_b"][:]
            y_vis = 10**ds["log10visc_b"][:]
            y_flx = ds["Fconv_b"][:]
        elif module == "spider":
            y_tmp = ds.get_dict_values(['data','temp_b'])
            y_phi = ds.get_dict_values(['data','phi_b'])
            y_vis = ds.get_dict_values(['data','visc_b'])
            y_flx = ds.get_dict_values(['data','Jconv_b'])

        # Store these data in the 2D array
        for j in range(nlev_b):
            arr_z1[i,j] = y_tmp[j]
            arr_z2[i,j] = y_phi[j]
            arr_z3[i,j] = y_vis[j]
            arr_z4[i,j] = y_flx[j]

    # Ensure that all values are float
    for a in (arr_z1,arr_z2,arr_z3,arr_z4,arr_yb):
        a = np.array(a,dtype=float)

    # Y-axis ticks
    yticks = np.linspace(np.amin(arr_yb),np.amax(arr_yb),4)
    yticks = [round(v) for v in yticks]
    for ax in axs:
        ax.set_ylim([yticks[-1], yticks[0]])
        ax.set_yticks(yticks)

    # Plot panel label
    panel_labels = ['(a)','(b)','(c)','(d)']
    for i in range(4):
        axs[i].text(0.015,0.97, panel_labels[i], transform=axs[i].transAxes,
                    verticalalignment="top", horizontalalignment="left", fontsize=12,
                    bbox=dict(boxstyle='square,pad=0.1',fc='white', alpha=0.4, linewidth=0))

    # Plot temperature
    cmap = cm.lajolla
    cax = make_axes_locatable(ax1).append_axes('right', size='5%', pad=0.05)
    norm = mpl.colors.Normalize(vmin=np.amin(arr_z1), vmax=np.amax(arr_z1))
    if use_contour:
        cf = ax1.contourf(times, arr_yb, arr_z1.T, cmap=cmap, norm=norm, levels=cblevels)
    else:
        cf = ax1.pcolormesh(times, arr_yb, arr_z1.T, cmap=cmap, norm=norm, rasterized=True)
    cb = fig.colorbar(cf, cax=cax, orientation='vertical')
    cb.set_label("Temperature [K]")
    cb.set_ticks([float(round(v, -2)) for v in np.linspace(np.amin(arr_z1), np.amax(arr_z1), numticks)])


    # Plot melt fraction
    cmap = cm.grayC
    cax = make_axes_locatable(ax2).append_axes('right', size='5%', pad=0.05)
    norm = mpl.colors.Normalize(vmin=0.0, vmax=1.0, clip=True)
    if use_contour:
        cf = ax2.contourf(times, arr_yb, arr_z2.T, cmap=cmap, norm=norm, levels=np.linspace(0.0, 1.0, cblevels))
    else:
        cf = ax2.pcolormesh(times, arr_yb, arr_z2.T, cmap=cmap, norm=norm, rasterized=True)
    cb = fig.colorbar(cf, cax=cax, orientation='vertical')
    cb.set_label("Melt fraction")
    cb.set_ticks(list(np.linspace(0.0,1.0,numticks)))

    # Plot viscosity
    cmap = cm.imola_r
    cax = make_axes_locatable(ax3).append_axes('right', size='5%', pad=0.05)
    if (np.amax(arr_z3) > 100.0*np.amin(arr_z3)):
        norm = mpl.colors.LogNorm(vmin=np.amin(arr_z3), vmax=np.amax(arr_z3))
    else:
        norm = mpl.colors.Normalize(vmin=np.amin(arr_z3), vmax=np.amax(arr_z3))
    if use_contour:
        cf = ax3.contourf(times, arr_yb, arr_z3.T, cmap=cmap, norm=norm, levels=cblevels)
    else:
        cf = ax3.pcolormesh(times, arr_yb, arr_z3.T, cmap=cmap, norm=norm, rasterized=True)
    cb = fig.colorbar(cf, cax=cax, orientation='vertical')
    cb.set_label("Viscosity [Pa s]")

    # Plot convective heat flux
    cmap = cm.acton
    cax = make_axes_locatable(ax4).append_axes('right', size='5%', pad=0.05)
    if (np.amax(arr_z4) > 100.0*np.amin(arr_z4)):
        norm = mpl.colors.SymLogNorm(vmin=np.amin(arr_z4), vmax=np.amax(arr_z4), linthresh=1.0)
    else:
        norm = mpl.colors.Normalize(vmin=np.amin(arr_z4), vmax=np.amax(arr_z4))
    if use_contour:
        cf = ax4.contourf(times, arr_yb, arr_z4.T, cmap=cmap, norm=norm, levels=cblevels)
    else:
        cf = ax4.pcolormesh(times, arr_yb, arr_z4.T, cmap=cmap, norm=norm, rasterized=True)
    cb = fig.colorbar(cf, cax=cax, orientation='vertical')
    cb.set_label("Convective flux [W m$^{-2}$]")
    cb.set_ticks([float(round(v, -1)) for v in np.linspace(np.amin(arr_z4), np.amax(arr_z4), numticks)])

    # Save plot
    ax4.set_xticks(np.linspace(times[0], times[-1], 5))
    fname = os.path.join(output_dir,"plot_interior_cmesh.%s"%plot_format)
    fig.subplots_adjust(top=0.98, bottom=0.07, right=0.85, left=0.13, hspace=0.11)
    fig.savefig(fname, dpi=200)


def plot_interior_cmesh_entry(handler: Proteus):

    # Which module was used?
    module = handler.config.interior.module
    if module == "spider":
        extension = ".json"
    elif module == "aragog":
        extension = "_int.nc"

    # Get data
    plot_times,_ = sample_output(handler, extension=extension, tmin=1e3, nsamp=99999)
    data = read_interior_data(handler.directories["output"], module, plot_times)

    # Plot fixed set from above
    plot_interior_cmesh(
        output_dir=handler.directories["output"],
        times=plot_times, data=data, module=module,
        plot_format=handler.config.params.out.plot_fmt,
    )


if __name__ == "__main__":
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_interior_cmesh_entry(handler)



# Variables and functions to help with offline chemistry plotting functions
# These do not do the plotting themselves
from __future__ import annotations

import glob
import json
import pickle as pkl

import numpy as np

from proteus.config import read_config


def offchem_read_year(output_dir, year_int, mx_clip_min=1e-30, mx_clip_max=1.0, read_const=False):
    """Read PT profile, VULCAN output file, and optionally runtime mixing ratios.

    Given the path to an offline chemistry output folder, containing year
    subfolders, this function reads data from a given year. It optionally
    reads the mixing ratios calculated by SPIDER at runtime, and also applies
    a clip to the mixing ratios.

    Parameters
    ----------
        output_dir : str
            Path to offline chemistry output folder
        year_int : int
            Year to be read from `output_dir`

        mx_clip_min=1e-30 : float
            Mixing ratio floor
        mx_clip_max=1.0 : float
            Mixing ratio ceiling
        read_const=False : bool
            Read the mixing ratios that were used during the PROTEUS simulation?

    Returns
    ----------
        year_data : dict
            Dictionary of output arrays for the given year.
    """

    year_data = {}  # Dictionary of mixing ratios, pressure, temperature

    # Read VULCAN output file
    with open(output_dir+"offchem/%d/output.vul" % year_int, "rb") as handle:
        vul_data = pkl.load(handle)


    # Save year
    year_data["year"] = int(year_int)

    # Parse mixing ratios
    vul_var = vul_data['variable']
    for si,sp in enumerate(vul_var['species']):
        year_data["mx_"+sp] = np.clip(np.array(vul_var['ymix'])[:,si],mx_clip_min,mx_clip_max)

    # Parse PT profile
    vul_atm = vul_data['atm']
    year_data["pressure"] =     np.array(vul_atm["pco"]) / 1e6  # convert to bar
    year_data["temperature"] =  np.array(vul_atm["Tco"])

    # Read melt-vapour eqm mixing ratios which were used to initialise VULCAN
    if read_const:
        vol_data_str = None
        with open(output_dir+"offchem/%d/vulcan_cfg.py" % year_int, "r") as f:
            lines = f.read().splitlines()
            for ln in lines:
                if "const_mix" in ln:
                    vol_data_str = ln.split("=")[1].replace("'",'"')
                    break

        if vol_data_str is None:
            raise Exception("Could not parse vulcan cfg file!")

        vol_data = json.loads(vol_data_str)
        for vol in vol_data:
            mv_mx = float(vol_data[vol])
            if mv_mx > 1e-12:
                year_data["mv_"+vol] = mv_mx

    return year_data

def offchem_read_grid(grid_dir):
    """Read-in all information from a GridOfflineChemistry output folder.

    Given the path to a grid folder, containing case_* subfolders, this
    function reads all of the data required to process and plot the grid output.

    Parameters
    ----------
        grid_dir : str
            Path to grid output folder

    Returns
    ----------
        years : np.ndarray
            Array of years in the grid [gpoints, nyears]
        opts : np.ndarray
            Array of OPTIONS dicts in the grid [gpoints]
        data : np.ndarray
            Array of ReadOfflineChemistry output dicts [gpoints, nyears]

    """

    # Get folders for each grid point
    point_folders = glob.glob(grid_dir + "/case_*/")
    gpoints = len(point_folders)

    if gpoints < 1:
        raise Exception("Unable to read offchem grid because no grid points were found in the output folder.")

    # Read grid point-by-point
    grid_years  = []  # For each gp, get years
    grid_opts   = []  # For each gp, read-in OPTIONS
    grid_data   = []  # For each gp, for each year, get offchem data
    for i in range(gpoints):
        fol         = point_folders[i]        # folder
        years_read  = glob.glob(fol + "/offchem/*/")  # years as file paths

        # years as integers
        def ytoint(ypath):
            return int(str(ypath).split("/")[-2])
        gp_years = sorted([ ytoint(y) for y in years_read ])
        grid_years.append(gp_years)

        # OPTIONS dictionary
        options = read_config(fol + "/init_coupler.toml")

        gp_opts = dict(options[0])
        grid_opts.append(gp_opts)

        # offchem data
        gp_data = []  # for each year in this gp
        for i,y in enumerate(gp_years):
            year_data = offchem_read_year(fol, y, read_const = True )
            gp_data.append(year_data)
        grid_data.append(gp_data)

    # Convert to numpy arrays and assert types
    years = np.array(grid_years, dtype=int)
    opts  = np.array(grid_opts,  dtype=dict)
    data  = np.array(grid_data,  dtype=dict)

    return years, opts, data

def offchem_slice_grid(years, opts, data, cvar_filter):
    """Slice the output of GridOfflineChemistry according to control variables.

    Provided data from the output of GridOfflineChemistry, this function
    extracts a 'cross-section' of the data according to the provided control-
    variable filter. It returns a subset of the total grid, in which every
    grid point matches the required control variable arguments.

    For example, if `cvar_filter = {"semimajoraxis": 0.1}`, then it would only
    return grid points which ran with a `semimajoraxis` of `0.1` exactly. All
    other grid points are excluded from the return value.

    Parameters
    ----------
        years : np.ndarray
            Array of years in the grid [gpoints, nyears]
        opts : np.ndarray
            Array of OPTIONS dicts in the grid [gpoints]
        data : np.ndarray
            Array of `offchem_read_year` output dicts [gpoints, nyears]
        cvar_filter : dict
            Dictionary of control variables

    Returns
    ----------
        slice_years : np.ndarray
            Sliced array of years in the grid [sgpoints, nyears]
        slice_opts : np.ndarray
            Sliced array of OPTIONS dicts in the grid [sgpoints]
        slice_data : np.ndarray
            Sliced array of ReadOfflineChemistry output dicts [sgpoints, nyears]
    """

    # Subgrid slice
    slice_years  = []
    slice_opts   = []
    slice_data   = []

    gpoints = len(opts)

    # For each grid point, for each year...
    # check their options to see they each match cvar_filter or not
    for gi in range(gpoints):

        gp_opts = opts[gi]

        exclude = False  # exclude this grid point from slice?

        # For each filter key
        for k in cvar_filter.keys():

            # Mismatching keys (this can just be ignored I think?)
            if k not in gp_opts.keys():
                print("WARNING: filter key '%s' is not present in OPTIONS" % k)
                continue

            # Does not match
            if cvar_filter[k] != gp_opts[k]:
                exclude = True
                break

        if not exclude:
            slice_years.append(years[gi])
            slice_opts.append(gp_opts)
            slice_data.append(data[gi])

    # Convert to numpy arrays and assert types
    slice_years = np.array(slice_years, dtype=int)
    slice_opts  = np.array(slice_opts,  dtype=dict)
    slice_data  = np.array(slice_data,  dtype=dict)

    if np.shape(slice_opts)[0] == 0:
        print("WARNING: No grid points left after slicing")

    return slice_years, slice_opts, slice_data



# Functions used to handle atmosphere temperature structure (running AGNI, etc.)
from __future__ import annotations

import glob
import logging
import os
from typing import TYPE_CHECKING

import numpy as np
from juliacall import Main as jl
from scipy.interpolate import PchipInterpolator

from proteus.atmos_clim.common import get_height_from_pressure, get_spfile_path
from proteus.utils.constants import gas_list
from proteus.utils.helper import UpdateStatusfile, create_tmp_folder, safe_rm
from proteus.utils.logs import GetCurrentLogfileIndex, GetLogfilePath

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

# Constant
AGNI_LOGFILE_NAME="agni_recent.log"

def sync_log_files(outdir:str):
    # Logfile paths
    agni_logpath = os.path.join(outdir, AGNI_LOGFILE_NAME)
    logpath = GetLogfilePath(outdir, GetCurrentLogfileIndex(outdir))

    # Copy logfile content
    with open(agni_logpath, "r") as infile:
        inlines = infile.readlines()

        with open(logpath, "a") as outfile:
            for i,line in enumerate(inlines):
                # First line of agni logfile has NULL chars at the start, for some reason
                if i == 0:
                    line = "[" + line.split("[", 1)[1]
                # copy the line
                outfile.write(line)

    # Remove logfile content
    with open(agni_logpath, "w") as hdl:
        hdl.write("")

def activate_julia(dirs:dict):

    log.info("Activating Julia environment")
    jl.seval("using Pkg")
    jl.Pkg.activate(dirs["agni"])

    # Plotting configuration
    jl.seval('ENV["GKSwstype"] = "100"')
    jl.seval("using Plots")
    jl.seval('default(label=nothing, dpi=250)')

    # Import AGNI
    jl.seval("using AGNI")

    # Setup logging from AGNI
    #    This handle will be kept open throughout the PROTEUS simulation, so the file
    #    should not be deleted at runtime. However, it will be emptied when appropriate.
    verbosity = 1
    logpath = os.path.join(dirs["output"], AGNI_LOGFILE_NAME)
    jl.AGNI.setup_logging(logpath, verbosity)

    log.debug("AGNI will log to '%s'"%logpath)


def _construct_voldict(hf_row:dict, dirs:dict):

    # get from hf_row
    vol_dict = {}
    vol_sum = 0.0
    for vol in gas_list:
        vol_dict[vol] = hf_row[vol+"_vmr"]
        vol_sum += vol_dict[vol]

    # Check that the total VMR is not zero
    if vol_sum < 1e-4:
        UpdateStatusfile(dirs, 20)
        raise Exception("All volatiles have a volume mixing ratio of zero")

    return vol_dict


def init_agni_atmos(dirs:dict, config:Config, hf_row:dict):
    """Initialise atmosphere struct for use by AGNI.

    Does not set the temperature profile.

    Parameters
    ----------
        dirs : dict
            Dictionary containing paths to directories
        config : Config
            Configuration options and other variables
        hf_row : dict
            Dictionary containing simulation variables for current iteration

    Returns
    ----------
        atmos : atmosphere.Atmos_t
            Atmosphere struct

    """

    log.debug("New AGNI atmosphere")

    atmos = jl.AGNI.atmosphere.Atmos_t()

    # Stellar spectrum path
    sflux_files = glob.glob(os.path.join(dirs["output"], "data", "*.sflux"))
    sflux_times = [ int(s.split("/")[-1].split(".")[0]) for s in sflux_files]
    sflux_path  = os.path.join(dirs["output"],
                                "data", "%d.sflux"%int(sorted(sflux_times)[-1]))

    # Spectral file path
    try_spfile = os.path.join(dirs["output"] , "runtime.sf")
    if os.path.exists(try_spfile):
        # exists => don't modify it
        input_sf =      try_spfile
        input_star =    ""
    else:
        # doesn't exist => AGNI will copy it + modify as required
        input_sf =      get_spfile_path(dirs["fwl"], config)
        input_star =    sflux_path

    # composition
    vol_dict = _construct_voldict(hf_row, dirs)

    # set condensation
    condensates = []
    if config.atmos_clim.agni.condensation:
        if len(vol_dict) == 1:
            # single-gas case
            condensates = list(vol_dict.keys())
        else:
            # get sorted gases (in order of decreasing VMR at the surface)
            vol_sorted = sorted(vol_dict.items(), key=lambda item: item[1])[::-1]

            # Set gases as condensates...
            condensates = ['H2O'] # always prefer H2O
            for v in vol_sorted:
                # add gas if it has non-zero abundance
                if (v[1] > 1e-30) and (v[0] not in condensates):
                    condensates.append(v[0])

            # Remove the least abundant gas from the list, so that we have something to
            #     fill the background with if everything else condenses at a given level
            condensates = condensates[:-1]

    # Chemistry
    chem_type = config.atmos_clim.agni.chemistry
    include_all = False
    fc_dir = "_unset"
    if chem_type == 'eq':
        # equilibrium
        include_all = True
        condensates = []

        # working folder for fastchem coupling
        fc_dir = create_tmp_folder()
        log.debug("Fastchem work folder: '%s'"%fc_dir)

    # Surface single-scattering albedo
    surface_material = config.atmos_clim.agni.surf_material
    if "greybody" in str(surface_material).lower():
        # Grey value
        surface_material = "greybody"
        log.debug("Using grey single-scattering surface properties")

    else:
        # Empirical values
        log.debug(f"Using '{surface_material}' single-scattering surface properties")
        surface_material = os.path.join(dirs["fwl"], surface_material)
        if not os.path.isfile(surface_material):
            raise FileNotFoundError(surface_material)

    # Setup struct
    jl.AGNI.atmosphere.setup_b(atmos,
                        dirs["agni"], dirs["output"], input_sf,

                        hf_row["F_ins"],
                        config.orbit.s0_factor,
                        config.atmos_clim.albedo_pl,
                        config.orbit.zenith_angle,

                        hf_row["T_surf"],
                        hf_row["gravity"], hf_row["R_int"],

                        int(config.atmos_clim.agni.num_levels),
                        hf_row["P_surf"],
                        config.atmos_clim.agni.p_top,

                        vol_dict, "",

                        flag_rayleigh = config.atmos_clim.rayleigh,
                        flag_cloud    = config.atmos_clim.cloud_enabled,
                        overlap_method  = config.atmos_clim.agni.overlap_method,

                        albedo_s=config.atmos_clim.surf_greyalbedo,
                        surface_material=surface_material,
                        condensates=condensates,
                        use_all_gases=include_all,
                        fastchem_work = fc_dir,

                        skin_d=config.atmos_clim.surface_d,
                        skin_k=config.atmos_clim.surface_k,
                        tmp_magma=hf_row["T_surf"], tmp_floor=config.atmos_clim.tmp_minimum
                        )

    # Allocate arrays
    jl.AGNI.atmosphere.allocate_b(atmos,input_star)

    # Set temperature profile from old NetCDF if it exists
    nc_files = glob.glob(os.path.join(dirs["output"],"data","*_atm.nc"))
    if len(nc_files) > 0:
        log.debug("Load NetCDF profile")

        nc_times = [ int(s.split("/")[-1].split("_")[0]) for s in nc_files]
        nc_path  = os.path.join(dirs["output"],
                                "data", "%d_atm.nc"%int(sorted(nc_times)[-1]))
        jl.AGNI.setpt.fromncdf_b(atmos, nc_path)

    # Otherwise, set to initial guess
    else:
        jl.AGNI.setpt.isothermal_b(atmos, 1200.0)
        # jl.AGNI.setpt.loglinear_b(atmos, min(900.0, hf_row["T_surf"]))

    # Logging
    sync_log_files(dirs["output"])

    return atmos


def deallocate_atmos(atmos):
    """
    Deallocate atmosphere struct
    """
    jl.AGNI.atmosphere.deallocate_b(atmos)
    safe_rm(str(atmos.fastchem_work))


def update_agni_atmos(atmos, hf_row:dict, dirs:dict):
    """Update atmosphere struct.

    Sets the new boundary conditions and composition.

    Parameters
    ----------
        atmos : atmosphere.Atmos_t
            Atmosphere struct
        hf_row : dict
            Dictionary containing simulation variables for current iteration
        dirs : dict
            Directories dictionary

    Returns
    ----------
        atmos : atmosphere.Atmos_t
            Atmosphere struct

    """

    # ---------------------
    # Update compositions
    vol_dict = _construct_voldict(hf_row, dirs)
    for g in vol_dict.keys():
        atmos.gas_vmr[g][:]  = vol_dict[g]
        atmos.gas_ovmr[g][:] = vol_dict[g]

    # ---------------------
    # Store old/current log-pressure vs temperature arrays
    p_old = list(atmos.p)
    t_old = list(atmos.tmp)
    nlev_c = len(p_old)

    #    extend to lower pressures
    p_old = [p_old[0]/10] + p_old
    t_old = [t_old[0]]    + t_old

    #    extend to higher pressures
    p_old = p_old + [p_old[-1]*10]
    t_old = t_old + [t_old[-1]]

    #    create interpolator
    itp = PchipInterpolator(np.log10(p_old), t_old)

    # ---------------------
    # Update surface pressure [Pa] and generate new grid
    atmos.p_boa = 1.0e5 * float(hf_row["P_surf"])
    jl.AGNI.atmosphere.generate_pgrid_b(atmos)

    # ---------------------
    # Update surface temperature(s)
    atmos.tmp_surf  = float(hf_row["T_surf"] )
    atmos.tmp_magma = float(hf_row["T_magma"])

    # ---------------------
    # Set temperatures at all levels
    for i in range(nlev_c):
        atmos.tmp[i]  = float( itp(np.log10(atmos.p[i]))  )
        atmos.tmpl[i] = float( itp(np.log10(atmos.pl[i])) )
    atmos.tmpl[-1]    = float( itp(np.log10(atmos.pl[-1])))

    # ---------------------
    # Update instellation flux
    atmos.instellation = float(hf_row["F_ins"])

    return atmos



def run_agni(atmos, loops_total:int, dirs:dict, config:Config, hf_row:dict):
    """Run AGNI atmosphere model.

    Calculates the temperature structure of the atmosphere and the fluxes, etc.
    Stores the new flux boundary condition to be provided to SPIDER.

    Parameters
    ----------
        atmos : atmosphere.Atmos_t
            Atmosphere struct
        loops_total : int
            Model total loops counter.
        dirs : dict
            Dictionary containing paths to directories
        config : Config
            Configuration options and other variables
        hf_row : dict
            Dictionary containing simulation variables for current iteration

    Returns
    ----------
        atmos : atmosphere.Atmos_t
            Atmosphere struct
        output : dict
            Output variables, as a dictionary

    """

    # Inform
    log.info("Running AGNI...")
    time_str = "%d"%hf_row["Time"]

    # atmosphere solver plotting frequency
    modplot = 0
    plot_jacobian = False
    if config.params.out.logging == "DEBUG":
        modplot = 1
        plot_jacobian = True

    # tracking
    agni_success = False  # success?
    attempts = 0          # number of attempts so far

    # make attempts
    while not agni_success:
        attempts += 1
        log.info("Attempt %d" % attempts)

        # default parameters
        linesearch  = 2
        easy_start  = False
        dx_max      = config.interior.spider.tsurf_atol+5.0
        ls_increase = 1.02
        perturb_all = False
        max_steps   = 100

        # first iteration parameters
        if loops_total == 0:
            linesearch  = 2
            easy_start  = True
            perturb_all = True
            dx_max      = 200.0
            ls_increase = 1.1
            max_steps   = 200

        # try different solver parameters if struggling
        if attempts == 2:
            linesearch  = 1
            dx_max     *= 2.0
            ls_increase = 1.1
            perturb_all = True

        log.debug("Solver parameters:")
        log.debug("    ls_method=%d, easy_start=%s, dx_max=%.1f, ls_increase=%.2f"%(
            linesearch, str(easy_start), dx_max, ls_increase
        ))

        # Try solving temperature profile
        agni_success = jl.AGNI.solver.solve_energy_b(atmos,
                            sol_type  = int(config.atmos_clim.surf_state_int),
                            method    = int(1),
                            chem_type = int(config.atmos_clim.agni.chemistry_int),

                            conduct=False, convect=True, sens_heat=True,
                            latent=config.atmos_clim.agni.condensation, rainout=True,

                            max_steps=int(max_steps), max_runtime=900.0,
                            conv_atol=float(config.atmos_clim.agni.solution_atol),
                            conv_rtol=float(config.atmos_clim.agni.solution_rtol),

                            ls_increase=float(ls_increase), ls_method=int(linesearch),
                            dx_max=float(dx_max), easy_start=easy_start,
                            perturb_all=perturb_all,

                            save_frames=False, modplot=int(modplot),
                            plot_jacobian=plot_jacobian
                            )

        # Move AGNI logfile content into PROTEUS logfile
        sync_log_files(dirs["output"])

        # Model status check
        if agni_success:
            # success
            log.info("Attempt %d succeeded" % attempts)
            break
        else:
            # failure
            log.warning("Attempt %d failed" % attempts)

            # Max attempts
            if attempts >= 2:
                log.error("Maximum attempts when executing AGNI")
                break

    # Write output data
    ncdf_path = os.path.join(dirs["output"],"data",time_str+"_atm.nc")
    jl.AGNI.dump.write_ncdf(atmos, ncdf_path)

    # Make plots
    if (config.params.out.plot_mod > 0) \
            and (loops_total % config.params.out.plot_mod == 0):

        fmt = config.params.out.plot_fmt
        jl.AGNI.plotting.plot_fluxes(atmos, os.path.join(dirs["output"],
                                                    "plot_fluxes_atmosphere.%s"%fmt))
        jl.AGNI.plotting.plot_vmr(atmos, os.path.join(dirs["output"],
                                                    "plot_vmr.%s"%fmt))
        jl.AGNI.plotting.plot_contfunc1(atmos, os.path.join(dirs["output"],
                                                    "plot_cff.%s"%fmt))

    # ---------------------------
    # Calculate observables
    # ---------------------------

    # observed height and derived bulk density
    jl.AGNI.atmosphere.calc_observed_rho_b(atmos)
    rho_obs = float(atmos.transspec_rho)
    p_obs   = float(atmos.transspec_p) # set by peak of contribution function
    z_obs   = float(atmos.transspec_r) - hf_row["R_int"] # transspec_r = R_int + z_obs

    # ---------------------------
    # Parse results
    # ---------------------------

    log.debug("Parse results")
    net_flux =      np.array(atmos.flux_n)
    LW_flux_up =    np.array(atmos.flux_u_lw)
    SW_flux_up =    np.array(atmos.flux_u_sw)
    SW_flux_down =  np.array(atmos.flux_d_sw)
    T_surf =        float(atmos.tmp_surf)

    # New flux from SOCRATES
    F_atm_new = net_flux[0]

    # Enforce positive limit on F_atm, if enabled
    if config.atmos_clim.prevent_warming:
        F_atm_lim = max( 1e-8 , F_atm_new )
    else:
        F_atm_lim = F_atm_new
    if not np.isclose(F_atm_lim , F_atm_new ):
        log.warning("Change in F_atm [W m-2] limited in this step!")
        log.warning("    %g  ->  %g" % (F_atm_new , F_atm_lim))

    log.info("SOCRATES fluxes (net@BOA, net@TOA, OLR): %.2e, %.2e, %.2e  W m-2" %
                                        (net_flux[-1], net_flux[0] ,LW_flux_up[0]))

    # XUV height in atm
    if config.escape.module == 'zephyrus':
        # escape level set by zephyrus config
        p_xuv = config.escape.zephyrus.Pxuv # [bar]
    else:
        # escape level set to surface
        p_xuv = hf_row["P_surf"] # [bar]
    p_xuv, z_xuv = get_height_from_pressure(atmos.p, atmos.z, p_xuv*1e5) # [Pa], [m]

    # final things to store
    output = {}
    output["F_atm"]  = F_atm_lim
    output["F_olr"]  = LW_flux_up[0]
    output["F_sct"]  = SW_flux_up[0]
    output["T_surf"] = T_surf
    output["p_obs"]  = p_obs/1e5 # convert [Pa] to [bar]
    output["z_obs"]  = z_obs
    output["rho_obs"]= rho_obs
    output["albedo"] = SW_flux_up[0]/SW_flux_down[0]
    output["p_xuv"]  = p_xuv/1e5        # Closest pressure from Pxuv    [bars]
    output["z_xuv"]  = z_xuv            # Height at Pxuv                [m]

    return atmos, output



# Contains routines for model termination and convergence

# Import utils-specific modules
from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

from proteus.utils.helper import UpdateStatusfile

if TYPE_CHECKING:
    from proteus import Proteus
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

# Summarise convergence criteria
def print_termination_criteria(config:Config):
    log.info("Active termination criteria")

    def _print_criterion(state, msg):
        log.info(f"    {msg:16} {'✔' if state else '✗'}  ")

    # Optionally enabled:
    _print_criterion(config.params.stop.solid.enabled,  "Solidification")
    _print_criterion(config.params.stop.radeqm.enabled, "Energy balance")
    _print_criterion(config.params.stop.escape.enabled, "Volatile escape")
    _print_criterion(config.params.stop.time.enabled,   "Maximum time")
    _print_criterion(config.params.stop.iters.enabled,  "Maximum loops")

    # Always enabled:
    _print_criterion(True, "Keepalive file")

# Print message in common format
def _msg_termination(msg:str):
    log.info("")
    log.info(f"===> {msg}! <===")
    log.info("")

# Solidified condition
def _check_solid(handler: Proteus) -> bool:
    log.debug("Check solidification")
    if handler.hf_row["Phi_global"] <= handler.config.params.stop.solid.phi_crit:
        UpdateStatusfile(handler.directories, 10)
        _msg_termination("Planet solidified!")
        return True
    return False

# Energy balance condition
def _check_radeqm(handler: Proteus) -> bool:
    log.debug("Check energy balance / radeqm")

    # Radiative equilibrium nominally occurs when F_atm is zero.
    # However, with tidal heating included, this is offset by the interior
    #   heat production of the planet. Energy balance is instead achieved
    #   when F_atm and F_tidal are equal.

    F_eps = handler.hf_row["F_atm"] - handler.hf_row["F_tidal"] - handler.hf_row["F_radio"]
    F_ref = abs(handler.hf_row["F_atm"]) * handler.config.params.stop.radeqm.rtol + handler.config.params.stop.radeqm.atol

    if abs(F_eps) <= F_ref:
        UpdateStatusfile(handler.directories, 14)
        _msg_termination("Planet has reached global energy balance")
        return True

     # Simulation when wants to warm up but cannot do so
    if handler.config.atmos_clim.prevent_warming and (F_eps < 0):
        UpdateStatusfile(handler.directories, 14)
        _msg_termination("Planet is no longer cooling")
        return True

    return False

# Volatiles have escaped
def _check_escape(handler: Proteus) -> bool:
    log.debug("Check escape")

    if handler.has_escaped \
        or (handler.hf_row["M_atm"] <= handler.config.params.stop.escape.mass_frac * handler.hf_all.iloc[0]["M_atm"]):

        UpdateStatusfile(handler.directories, 15)
        _msg_termination("Atmosphere has escaped")
        return True
    return False

# Maximum time
def _check_time(handler: Proteus) -> bool:
    log.debug("Check maximum time")

    if handler.hf_row["Time"] >= handler.config.params.stop.time.maximum:
        UpdateStatusfile(handler.directories, 13)
        _msg_termination("Target time reached")
        return True
    return False

# Maximum iterations
def _check_maxiter(handler: Proteus) -> bool:
    log.debug("Check maximum iterations")

    if handler.loops["total"] > handler.loops["total_loops"]:
        UpdateStatusfile(handler.directories, 12)
        _msg_termination("Maximum number of iterations reached")
        return True
    return False

# Minimum iterations (return true when exit is allowed)
def _check_miniter(handler: Proteus, finished:bool) -> bool:
    log.debug("Check minimum iterations")

    # Number of iterations not reached
    if handler.loops["total"] <= handler.loops["total_min"]:
        if finished:
            # Model thinks that it is done
            log.warning("Minimum number of iterations not yet attained; continuing...")
            UpdateStatusfile(handler.directories, 1)
        return False # do not exit
    return True

def _check_keepalive(handler: Proteus) -> bool:
    log.debug("Check keepalive file")

    if not os.path.exists(handler.lockfile):
        UpdateStatusfile(handler.directories, 25)
        _msg_termination("Model exit was requested")
        return True
    return False

def check_termination(handler: Proteus) -> bool:
    """
    Decide if the model should stop or not.

    Updates the status file. Does not modify any variables in the handler.

    Parameters:
    --------------
        handler: Proteus
            Proteus instance

    Returns:
    --------------
        fininshed: bool
            Should the model terminate now?
    """

    # Quantify model state
    finished = False

    # Stop simulation when planet is completely solidified
    if handler.config.params.stop.solid.enabled:
        finished = finished or _check_solid(handler)

    # Stop simulation when at global energy balance
    if handler.config.params.stop.radeqm.enabled:
        finished = finished or _check_radeqm(handler)

    # Atmosphere has escaped
    if handler.config.params.stop.escape.enabled:
        finished = finished or _check_escape(handler)

    # Maximum time reached
    if handler.config.params.stop.time.enabled:
        finished = finished or _check_time(handler)

    # Maximum loops reached
    if handler.config.params.stop.iters.enabled:
        finished = finished or _check_maxiter(handler)

    # Minimum loops reached
    if handler.config.params.stop.iters.enabled:
        finished = finished and _check_miniter(handler, finished)

    # Check if keepalive file has been removed
    #    This means that the model should exit ASAP.
    #    This check should always come last.
    finished = finished or _check_keepalive(handler)

    # Set state
    return finished



# Function used to run VULCAN in online mode
from __future__ import annotations

import os
import pickle as pkl
import re
import shutil
import subprocess
from typing import TYPE_CHECKING

import numpy as np

from proteus.utils.constants import element_list, gas_list

if TYPE_CHECKING:
    from proteus.config import Config


def RunVULCAN( atm, time, loop_counter, dirs, runtime_helpfile, config: Config ):
    '''
    Run online atmospheric chemistry via VULCAN kinetics

    This function is deprecated.
    '''

    # Runtime info
    print("VULCAN run... (loop =", loop_counter, ")")

    # Copy template file
    vul_cfg = dirs["vulcan"]+"vulcan_cfg.py"                    # Template configuration file for VULCAN
    if os.path.exists(vul_cfg):
        os.remove(vul_cfg)
    shutil.copyfile(dirs["utils"]+"init_vulcan_cfg.py",vul_cfg)

    # Delete old PT profile
    vul_ptp = dirs["vulcan"]+"output/PROTEUS_PT_input.txt"      # Path to input PT profile
    if os.path.exists(vul_ptp):
        os.remove(vul_ptp)


    hf_recent = runtime_helpfile.iloc[-1]

    # Write missing parameters to cfg file
    with open(vul_cfg, 'a') as vcf:

        vcf.write("# < PROTEUS INSERT > \n")

        # System/planet parameters
        vcf.write("Rp = %1.5e \n"           % (atm.planet_radius*100.0))        # Radius [cm]
        vcf.write("gs = %g \n"              % (atm.grav_s*100.0))               # Surface gravity [cm/s^2]
        vcf.write("sl_angle = %g \n"        % (atm.zenith_angle*3.141/180.0))   # Solar zenith angle [rad]
        vcf.write("orbit_radius = %1.5e \n" % config["semimajoraxis"]) # Semi major axis [AU]
        vcf.write("r_star = %1.5e \n"       % config["star_radius"])   # Star's radius [R_sun]

        # Set background gas based on gas with highest mixing ratio from list of options
        bg_gas = np.array([ 'H2', 'N2', 'O2', 'CO2' ])
        bg_val = np.array([ hf_recent["%s_mr"%gas] for gas in bg_gas ])  # Get values
        bg_mask = np.argsort(bg_val)  # Get sorting mask
        bg_val = bg_val[bg_mask]   # Sort gas values
        bg_gas = bg_gas[bg_mask]   # Sort gas names
        if (bg_val[-1] < 2.0*bg_val[-2]) or (bg_val[-1] < 0.5):
            print("Warning: Background gas '%s' is not significantly abundant!"%bg_gas[-1])
            print("         Mixing ratio of '%s' is %g" % (bg_gas[-1],bg_val[-1]))
            print("         Mixing ratio of '%s' is %g" % (bg_gas[-2],bg_val[-2]))
        vcf.write("atm_base = '%s' \n"      % str(bg_gas[-1]))

        # Pressure grid limits
        vcf.write("P_b = %1.5e \n"          % float(hf_recent["P_surf"]*1.0e6))        # pressure at the bottom (dyne/cm^2)
        vcf.write("P_t = %1.5e \n"          % float(config["P_top"]*1.0e6))   # pressure at the top (dyne/cm^2)

        # Plotting behaviour
        vcf.write("use_live_plot  = %s \n"  % str(bool(config.params.out.plot_mod > 0)))

        # Make copy of element_list as a set, since it'll be used a lot in the code below
        set_elem_list = set(element_list)

        # Rayleigh scattering gases
        rayleigh_candidates = ['N2','O2', 'H2']
        rayleigh_str = ""
        for rc in rayleigh_candidates:
            if set(re.sub('[1-9]', '', rc)).issubset(set_elem_list):  # Remove candidates which aren't supported by elem_list
                rayleigh_str += "'%s',"%rc
        rayleigh_str = rayleigh_str[:-1]
        vcf.write("scat_sp = [%s] \n" % rayleigh_str)

        # Gases for diffusion-limit escape at TOA
        # escape_candidates = ['H2','H']
        # escape_str = ""
        # for ec in escape_candidates:
        #     if set(re.sub('[1-9]', '', ec)).issubset(set_elem_list):  # Remove candidates which aren't supported by elem_list
        #         escape_str += "'%s',"%ec
        # escape_str = escape_str[:-1]
        # vcf.write("diff_esc = [%s] \n" % escape_str)

        # Atom list
        atom_str = ""
        for elem in element_list:
            atom_str += "'%s',"%elem
        atom_str = atom_str[:-1]
        vcf.write("atom_list  = [%s] \n" % atom_str)

        # Choose most appropriate chemical network and species-to-plot
        oxidising = False

        for inert in ['He','Xe','Ar','Ne','Kr']:  # Remove inert elements from list, since they don't matter for reactions
            set_elem_list.discard(inert)

        if set_elem_list == {'H','C','O'}:
            net_str = 'thermo/CHO_photo_network.txt'
            plt_spe = ['H2', 'H', 'H2O', 'C2H2', 'CH4', 'CO2']

        elif set_elem_list == {'N','H','C','O'}:
            if oxidising:
                net_str = 'thermo/NCHO_full_photo_network.txt'
                plt_spe = ['N2', 'O2', 'H2', 'H2O', 'NH3', 'CH4', 'CO', 'O3']
            else:
                net_str = 'thermo/NCHO_photo_network.txt'
                plt_spe = ['H2', 'H', 'H2O', 'OH', 'CH4', 'HCN', 'N2', 'NH3']

        elif set_elem_list == {'S','N','H','C','O'}:
            if oxidising:
                net_str = 'thermo/SNCHO_full_photo_network.txt'
                plt_spe = ['O2', 'N2', 'O3', 'H2', 'H2O', 'NH3', 'CH4', 'SO2', 'S']
            else:
                net_str = 'thermo/SNCHO_photo_network.txt'
                plt_spe = ['N2', 'H2', 'S', 'H', 'OH', 'NH3', 'CH4', 'HCN']

        vcf.write("network  = '%s' \n" % net_str)
        plt_str = ""
        for spe in plt_spe:
            plt_str += "'%s'," % spe
        vcf.write("plot_spec = [%s] \n" % plt_str[:-1])

        # Bottom boundary mixing ratios are fixed according to SPIDER (??)
        # fix_bb_mr = "{"
        # for v in gas_list:
        #     fix_bb_mr += " '%s' : %1.5e ," % (v,hf_recent["%s_mr"%v])
        # fix_bb_mr = fix_bb_mr[:-1]+" }"
        # vcf.write("use_fix_sp_bot = %s \n"   % str(fix_bb_mr))
        vcf.write("use_fix_sp_bot = {} \n")


        # Has NOT run atmosphere before
        if (loop_counter["atm"] == 0):

            # PT profile
            # vcf.write("atm_type = 'isothermal' \n")
            # vcf.write("Tiso = %g \n"        % float(OPTIONS["T_eqm"]))

            # Abundances
            vcf.write("ini_mix = 'const_mix' \n")  # other options: 'EQ', 'table', 'vulcan_ini'
            const_mix = "{"
            for v in gas_list:
                const_mix += " '%s' : %1.5e ," % (v,hf_recent["%s_mr"%v])
            const_mix = const_mix[:-1]+" }"
            vcf.write("const_mix = %s \n"   % str(const_mix))

        # Has run atmosphere before
        # else:

            # PT Profile
            # vcf.write("atm_type = 'file' \n")
            # vcf.write("atm_file = '%s' \n" % vul_ptp)

            # Abundances
            # vcf.write("ini_mix = 'vulcan_ini' \n")
            # vcf.write("vul_ini = 'output/PROTEUS_MX_input.vul' \n")
            # ! WRITE ABUNDANCES HERE


        vcf.write("# </ PROTEUS INSERT > \n")
        vcf.write(" ")

    # Write PT profile (to VULCAN, from JANUS)
    vul_PT = np.array(
        [np.array(atm.pl)  [::-1] * 10.0,
         np.array(atm.tmpl)[::-1]
        ]
    ).T
    header = "#(dyne/cm2)\t (K) \n Pressure\t Temp"
    np.savetxt(vul_ptp,vul_PT,delimiter="\t",header=header,comments='',fmt="%1.5e")

    # Write spectrum (using most recent output spectrum from $COUPLERDIR/output)
    # Spectra in output are scaled to 1 AU, but VULCAN wants it at the star's surface.
    print("")
    print("WARNING: VULCAN IS NOT YET USING THE CORRECT SPECTRUM!")
    print("")


    # Switch to VULCAN directory, run VULCAN, switch back to main directory
    vulcan_run_cmd = "python vulcan.py"
    if (loop_counter["atm"] > 0):      # If not first run, skip building chem_funcs
        vulcan_run_cmd += " -n"

    vulcan_print = open(dirs["output"]+"vulcan_recent.log",'w')

    os.chdir(dirs["vulcan"])
    subprocess.run([vulcan_run_cmd], shell=True, check=True, stdout=vulcan_print)
    os.chdir(dirs["proteus"])

    vulcan_print.close()

    # Copy VULCAN output data file to output folder
    vulcan_recent = dirs["output"]+str(int(time))+"_atm_chemistry.vul"
    shutil.copyfile(dirs["vulcan"]+'output/PROTEUS_MX_output.vul', vulcan_recent )

    # Read in data from VULCAN output
    with (open(vulcan_recent, "rb")) as vof:
        vul_data = pkl.load(vof)

    print(vul_data)

    return atm



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from proteus.utils.constants import M_earth, M_jupiter, R_earth, R_jupiter

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)

EARTH_COLOUR="hotpink"

def _get_exo_data(fwl_dir:str):
    popfile = os.path.join(fwl_dir, "planet_reference", "Exoplanets", "DACE_PlanetS.csv")
    return pd.read_csv(popfile,comment="#")

def _get_mr_data(fwl_dir:str):
    z19 = os.path.join(fwl_dir, "mass_radius", "Mass-radius", "Zeng2019")

    # Set paths
    curves = {
        "Earth-like":       os.path.join(z19, "massradiusEarthlikeRocky.txt"),
        r"MgSiO3":          os.path.join(z19, "massradiusmgsio3.txt"),
        r"H$_2$O, 300 K":   os.path.join(z19, "massradius_100percentH2O_300K_1mbar.txt"),
        r"H$_2$O, 1000 K":  os.path.join(z19, "massradius_100percentH2O_1000K_1mbar.txt"),
    }

    # Replace paths with the data
    for k in curves.keys():
        data = np.loadtxt(curves[k]).T
        mask = np.argsort(data[0])
        curves[k] = [data[0][mask],data[1][mask]]

    return curves

def plot_population_mass_radius(hf_all:pd.DataFrame, output_dir: str, fwl_dir:str,
                                    plot_format:str, t0: float=100.0,
                                    show_errors:bool = False, m_max:float = 10.0):

    """
    Plot planetary evolution on a population mass-radius diagram.
    """

    # Get values from simulation
    hf_crop = hf_all.loc[hf_all["Time"]>t0]
    time = np.array(hf_crop["Time"])
    if len(time) < 3:
        log.warning("Cannot make plot with less than 3 samples")
        return

    log.info("Plot population (mass-radius)")
    sim_rad = np.array(hf_crop["z_obs"]+hf_crop["R_int"])/ R_earth
    sim_mas = np.array(hf_crop["M_planet"]) / M_earth

    # Get exoplanet values from database
    exo = _get_exo_data(fwl_dir)
    exo = exo.loc[exo["Planet Mass [Mjup]"] * M_jupiter / M_earth <= m_max*1.1]

    exo_mas_val = exo["Planet Mass [Mjup]"]             * M_jupiter / M_earth
    exo_mas_upp = exo["Planet Mass - Upper Unc [Mjup]"] * M_jupiter / M_earth
    exo_mas_low = exo["Planet Mass - Lower Unc [Mjup]"] * M_jupiter / M_earth

    exo_rad_val = exo["Planet Radius [Rjup]"]             * R_jupiter / R_earth
    exo_rad_upp = exo["Planet Radius - Upper Unc [Rjup]"] * R_jupiter / R_earth
    exo_rad_low = exo["Planet Radius - Lower Unc [Rjup]"] * R_jupiter / R_earth

    # Get Mass-Radius curves from files
    mrdata = _get_mr_data(fwl_dir)

    # Create plot
    scale = 1.0
    fig,ax = plt.subplots(1,1,figsize=(6*scale,5*scale))

    # M-R curves
    for k in mrdata.keys():
        ax.plot(mrdata[k][0],mrdata[k][1], label=k, linestyle='dashed', alpha=0.8, zorder=0)

    # Exoplanets
    lbl = "Exoplanets"
    col = 'grey'
    ms=6
    if show_errors:
        ax.errorbar(exo_mas_val, exo_rad_val,
                    xerr=[exo_mas_low, exo_mas_upp],
                    yerr=[exo_rad_low, exo_rad_upp],
                    label=lbl, ls='none', ms=ms, color=col, elinewidth=0.9, zorder=1)
    else:
        ax.scatter(exo_mas_val, exo_rad_val, color=col, label=lbl, s=ms, zorder=1)

    # Solar system planets
    ax.scatter(1.0,  1.0, label="Modern Earth", color=EARTH_COLOUR, zorder=2)

    # Simulation
    col='k'
    ax.plot(sim_mas, sim_rad, label="Simulation", color=col, lw=1.5, zorder=98)
    ax.scatter(sim_mas[-1], sim_rad[-1], color=col, s=ms, zorder=99)

    # Save figure
    ax.set_ylabel(r"Planet radius [$R_{\oplus}$]")
    ax.set_ylim(bottom=0, top=max(np.amax(exo_rad_val), np.amax(sim_rad)))

    ax.set_xlabel(r"Planet mass [$M_{\oplus}$]")
    ax.set_xlim(left=0,right=m_max)

    ax.legend(loc='upper left', ncol=2)
    plt.close()
    plt.ioff()
    fig.savefig(output_dir+"/plot_population_mass_radius."+plot_format,
                bbox_inches='tight', dpi=200)


def plot_population_time_density(hf_all:pd.DataFrame, output_dir: str, fwl_dir:str,
                                 plot_format:str, t0: float=100.0,
                                 xmin:float=1e5, ymin:float=0.1, ymax:float=10.0):
    """
    Plot planetary evolution on population time-density diagram
    """

    # Get values from simulation
    hf_crop = hf_all.loc[hf_all["Time"]>t0]
    time = np.array(hf_crop["Time"])
    if len(time) < 3:
        log.warning("Cannot make plot with less than 3 samples")
        return

    log.info("Plot population (time-density)")
    sim_rad = np.array(hf_crop["z_obs"]+hf_crop["R_int"])
    sim_mas = np.array(hf_crop["M_planet"])
    sim_rho = 3*sim_mas/(4*np.pi*sim_rad**3)  * 0.001

    # Get values from database
    exo = _get_exo_data(fwl_dir)
    exo_age_val = np.array(exo["Stellar Age [Gyr]"]   )  * 1e9    # yr
    exo_rho_val = np.array(exo["Planet Density [g/cm**3] - Computation"]) # g cm-3
    xmax = max(np.nanmax(exo_age_val), np.amax(time))

    # Make sure that we actually plot the simulation data
    if xmin > np.amax(time):
        xmin = time[0]
    ymin = min(ymin, np.amin(sim_rho[2:])/2)
    ymax = max(ymin, np.amax(sim_rho[2:])*2)

    # Create plot
    scale = 1.0
    lw=1.8
    fig,ax = plt.subplots(1,1,figsize=(6*scale,5*scale))

    # Reference values
    ax.axhline(y=1.0, label=r"H$_2$O", color='tab:blue', alpha=0.8, zorder=0, ls='dashed')

    # Exoplanets
    lbl = "Exoplanets"
    col = 'grey'
    ms=8
    ax.scatter(exo_age_val, exo_rho_val, color=col, label=lbl, s=ms, zorder=1)

    # Modern Earth
    ax.scatter(4.6e9, 5513*0.001, label="Modern Earth", color=EARTH_COLOUR, zorder=3)

    # Simulation
    col='k'
    ax.plot([time[-1],1e13],[sim_rho[-1],sim_rho[-1]], color=col, lw=lw, ls='dotted', label="Extrapolated", zorder=97)
    ax.plot(time, sim_rho, label="Simulation", color=col, lw=lw, zorder=98)
    ax.scatter(time[-1], sim_rho[-1], color=col, s=ms, zorder=99)


    # Save figure
    ax.set_ylabel(r"Bulk density [g cm$^{-3}$]")
    ax.set_yscale("log")
    ax.set_ylim(bottom=ymin, top=ymax)

    ax.set_xlabel(r"Time [yr]")
    ax.set_xscale("log")
    ax.set_xlim(left=xmin,right=xmax*1.2)

    ax.legend()
    plt.close()
    plt.ioff()
    fig.savefig(output_dir+"/plot_population_time_density."+plot_format,
                bbox_inches='tight', dpi=200)


def plot_population_entry(handler: Proteus):
    # read helpfile
    hf_all = pd.read_csv(os.path.join(handler.directories['output'], "runtime_helpfile.csv"), sep=r"\s+")

    # make plots
    plot_population_mass_radius(
        hf_all=hf_all,
        output_dir=handler.directories["output"],
        fwl_dir=handler.directories["fwl"],
        plot_format=handler.config.params.out.plot_fmt
    )

    plot_population_time_density(
        hf_all=hf_all,
        output_dir=handler.directories["output"],
        fwl_dir=handler.directories["fwl"],
        plot_format=handler.config.params.out.plot_fmt
    )

if __name__ == '__main__':
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_population_entry(handler)



# Physical, numerical, etc constants

# Avoid importing from any of the files within utils/

# Astronomical constants
from __future__ import annotations

L_sun           = 3.828e+26             # W, IAU definition
R_sun           = 6.957e8               # m
R_sun_cm        = 100 * R_sun           # cm
M_sun           = 1.988416e30           # kg
AU              = 1.495978707e+11       # m
AU_cm           = AU * 100.0            # cm
R_gas           = 8.31446261815324      # J K−1 mol−1
M_earth         = 5.972E24              # kg
R_earth         = 6.335439e6            # m
R_core_earth    = 3485000.0             # m
M_core_earth    = 1.94E24               # kg
mol             = 6.02214076e+23        # mol definition

# Earth heat flux, globally averaged [W m-2]
# https://se.copernicus.org/articles/1/5/2010/
F_earth         = 47.2e12 / (4 * 3.141 * R_earth * R_earth)

# Io radius [m] and tidal heat flux [W m-2]
R_io            = 1.82149e6     # https://ssd.jpl.nasa.gov/sats/phys_par/#refs
F_io            = 2.24          # https://www.nature.com/articles/nature08108

# https://nssdc.gsfc.nasa.gov/planetary/factsheet/
M_jupiter       = 1.898e27              # kg
R_jupiter       = 1.42984e8 / 2         # m

# Time constants
secs_per_day    = 24 * 60 * 60
secs_per_year   = secs_per_day * 365.25

# Values from Pierrehumbert (2010) workbook
const_h = 6.626075540e-34    #Planck's constant
const_c = 2.99792458e8       #Speed of light
const_k =1.38065812e-23      #Boltzman thermodynamic constant
const_sigma = 5.67051196e-8  #Stefan-Boltzman constant
const_G = 6.67428e-11        #Gravitational constant (2006 measurements)
const_N_avogadro = 6.022136736e23  #Avogadro's number
const_R_gas = 8.31446261815324 # Universal gas constant, J.K-1.mol-1

# Supported gases
vol_list = ["H2O", "CO2", "H2", "CH4", "CO", "N2", "S2", "SO2"]
vap_list = ["SiO", "SiO2", "MgO", "FeO2"]
gas_list = vol_list + vap_list

# Supported elements
element_list = ["H", "O", "C", "N", "S", "Si", "Mg", "Fe", "Na"]

## Constant from Zephyrus
ergcm2stoWm2  = 1e-3                  # convert [erg s-1 cm-2] to [W m-2]
s2yr          = 1/(3600*24*365)       # convert [seconds]      to [years]
# Sun parameters
Rs      = 6.957e8                      # Solar radius                          [m]
Ms      = 1.98847e30                   # Solar mass                            [kg]
age_sun = 4.603e9                      # Age of the Sun                        [yr]
# Earth parameters
Me_atm            = 5.15e18            # Mass of the Earth atmopshere          [kg]
Fxuv_earth_10Myr  = 14.67              # Fxuv received on Earth at t = 10 Myr -> see Fig 9. Wordsworth+18 [W m-2]
Fxuv_earth_today  = 4.64e-3            # Stellar flux received on Earth today  [W m-2]
age_earth         = 4.543e9            # Age of the Earth                      [yr]
e_earth           = 0.0167             # Earth eccentricity                    [dimensionless]
a_earth           = 1                  # Earth semi-major axis                 [au]

# Radionuclide values below are from Ruedas (2017), via SPIDER
# Natural concentrations (provided in config file) can
#      be obtained from Turcotte & Schubert, 2014, p. 170
radnuc_data = {
    "k40": {
        "abundance": 1.1668E-4, # 40K/K
        "heatprod":  2.8761E-5, # W/kg
        "halflife":  1.248e9    # yr
    },

    "th232": {
        "abundance": 1.0,       # 232Th/Th
        "heatprod":  2.6368E-5, # W/kg
        "halflife":  14e9       # yr
    },

    "u235": {
        "abundance": 0.0072045,  # 235U/U
        "heatprod":  5.68402E-4, # W/kg
        "halflife":  0.704e9     # yr
    },

    "u238": {
        "abundance": 0.9927955, # 40K/K
        "heatprod":  9.4946E-5, # W/kg
        "halflife":  4.468e9    # yr
    },
}



from __future__ import annotations

import logging

from attrs import define, field
from attrs.validators import ge, gt, in_, le

from ._converters import none_if_none

log = logging.getLogger('fwl.' + __name__)


def tmp_max_bigger_than_tmp_min(instance, attribute, value):
    if value <= instance.tmp_minimum:
        raise ValueError("'tmp_maximum' has to be bigger than 'tmp_minimum'.")


def warn_if_dummy(instance, attribute, value):
    if (instance.module == 'dummy') and value:
        raise ValueError('Dummy atmos_clim module is incompatible with Rayleigh scattering')

def check_overlap(instance, attribute, value):
    _overlaps = ("ro", "ee", "rorr")
    if value not in _overlaps:
        raise ValueError("Overlap type must be one of " + str(_overlaps))

@define
class AtmosClim:
    """Atmosphere parameters, model selection.

    Attributes
    ----------
    prevent_warming: bool
        When True, require the planet to monotonically cool over time.
    surface_d: float
        Conductive skin thickness [m],
    surface_k: float
        Conductive skin thermal conductivity [W m-1 K-1].
    cloud_enabled: bool
        Enable water cloud radiative effects.
    cloud_alpha: float
        Condensate retention fraction (0 => full rainout, 1 => fully retained).
    surf_state: str
        Surface energy balance scheme. Choices: "mixed_layer", "fixed", "skin".
    surf_greyalbedo : float
        Grey surface albedo.
    albedo_pl: float
        Planetary/bold albedo used to emulate scattering [dimensionless].
    rayleigh: bool
        Include Rayleigh scattering in the radiative transfer calculations.
    tmp_minimum: float
        Minimum temperature throughout the atmosphere [K].
    tmp_maximum: float
        Maximum temperature throughout the atmosphere [K].
    module: str
        Which atmosphere module to use.
    agni: Agni
        Config parameters for AGNI atmosphere module
    janus: Janus
        Config parameters for JANUS atmosphere module
    dummy: Dummy
        Config parameters for dummy atmosphere module
    """

    prevent_warming: bool
    surface_d: float = field(validator=gt(0))
    surface_k: float = field(validator=gt(0))
    cloud_enabled: bool
    cloud_alpha: float = field(validator=(ge(0), le(1)))
    surf_state: str = field(validator=in_(('mixed_layer', 'fixed', 'skin')))
    surf_greyalbedo:float = field(validator=(ge(0),le(1)))
    albedo_pl: float = field(validator=(ge(0), le(1)))
    rayleigh: bool = field(validator=warn_if_dummy)
    tmp_minimum: float = field(validator=gt(0))
    tmp_maximum: float = field(validator=tmp_max_bigger_than_tmp_min)

    module: str = field(validator=in_(('dummy', 'agni', 'janus')))

    agni: Agni
    janus: Janus
    dummy: Dummy

    @property
    def surf_state_int(self) -> int:
        """Return integer state for agni."""
        return ('mixed_layer', 'fixed', 'skin').index(self.surf_state)


@define
class Agni:
    """AGNI atmosphere module.

    Attributes
    ----------
    p_top: float
        Top of atmosphere grid pressure [bar].
    spectral_group: str
        Spectral file codename defining the gas opacities to be included. See [documentation](https://raw.githubusercontent.com/FormingWorlds/PROTEUS/main/docs/assets/spectral_files.pdf).
    spectral_bands: str
        Number of wavenumer bands in k-table. See documentation.
    surf_material : str
        File name for material used to set surface single-scattering properties, relative to FWL data directory. Set to 'greybody' to use `surf_greyalbedo`. See [documentation](https://fwl-proteus.readthedocs.io/en/latest/data/#surfaces) for potential options.
    num_levels: str
        Number of atmospheric grid levels.
    chemistry: str | None
        Treatment of self-consistent atmospheric chemsitry. Choices: "none", "eq".
    solution_atol: float
        Absolute tolerance on the atmosphere solution.
    solution_rtol: float
        Relative tolerance on the atmosphere solution.
    overlap_method: str
        Gas overlap method. Choices: random overlap ("ro"), RO with resorting+rebinning ("rorr"), equivalent extinction ("ee").
    condensation: bool
        Enable volatile condensation/phase change in the atmosphere.
    """

    p_top: float = field(validator=gt(0))
    spectral_group: str
    spectral_bands: str
    surf_material: str
    num_levels: int = field(validator=ge(15))
    chemistry: str = field(validator=in_((None, "eq")), converter=none_if_none)
    solution_atol: float = field(validator=gt(0))
    solution_rtol: float = field(validator=gt(0))
    overlap_method: str = field(validator=check_overlap)
    condensation: bool

    @property
    def chemistry_int(self) -> int:
        """Return integer state for agni."""
        return 1 if self.chemistry else 0

@define
class Janus:
    """JANUS atmosphere module.

    Attributes
    ----------
    p_top: float
        Top of atmosphere grid pressure [bar].
    p_obs: float
        Pressure level probed by observations [bar]
    spectral_group: str
        Spectral file codename defining the gas opacities to be included. See [documentation](https://raw.githubusercontent.com/FormingWorlds/PROTEUS/main/docs/assets/spectral_files.pdf).
    spectral_bands: str
        Number of wavenumer bands in k-table. See documentation.
    F_atm_bc: int
        Measure outgoing flux using value at TOA (0) or surface (1).
    num_levels: int
        Number of atmospheric grid levels.
    tropopause: str | None
        Scheme for determining tropopause location. Choices: "none", "skin", "dynamic".
    overlap_method: str
        Gas overlap method. Choices: random overlap ("ro"), RO with resorting+rebinning ("rorr"), equivalent extinction ("ee").
    """

    p_top: float = field(validator=gt(0))
    p_obs: float = field(validator=gt(0))
    spectral_group: str
    spectral_bands: str
    F_atm_bc: int = field(validator=in_((0, 1)))
    num_levels: int = field(validator=ge(15))
    tropopause: str | None = field(
        validator=in_((None, 'skin', 'dynamic')), converter=none_if_none
    )
    overlap_method: str = field(validator=check_overlap)

@define
class Dummy:
    """Dummy atmosphere module.

    Attributes
    ----------
    gamma: float
        Atmosphere opacity between 0 and 1.
    """

    gamma: float



# Dummy star module
from __future__ import annotations

import logging

import numpy as np

from proteus.utils.constants import AU, const_sigma
from proteus.utils.phys import planck_wav

log = logging.getLogger("fwl."+__name__)

PLANCK_MIN_TEMPERATURE:float = 0.1

def generate_spectrum(tmp:float, R_star:float):
    '''
    Get stellar spectrum at 1 AU, assuming that the star emits like a blackbody.

    Parameters
    -----------
        tmp : float
            Temperature [K]
        R_star : float
            Stellar radius [m]

    Returns
    -----------
        wl_arr : list
            Wavelength bin centres [nm]
        fl_arr : list
            Stellar spectral flux density at 1 AU from star [erg s-1 cm-2 nm-1]
    '''

    log.debug("Generating stellar spectrum at Teff=%.0f K"%tmp)

    # Allocate wavelength array
    wl_min = 1e-10  # 0.1 nm
    wl_max = 1e-1   # 10 cm
    wl_pts = 600    # number of points to sample
    wl_arr = np.logspace(np.log10(wl_min), np.log10(wl_max), wl_pts)

    # Allocate flux array with zeros
    fl_arr = np.zeros(np.shape(wl_arr))

    # If Teff=0, keep fluxes at zero. This is equivalent to turning the star 'off'.

    # Else, for non-zero effective temperatures...
    if tmp > PLANCK_MIN_TEMPERATURE:

        # Evaluate planck function in each bin
        for i,wav in enumerate(wl_arr):
            fl_arr[i] = planck_wav(tmp, wav) # W m-2 m-1 at stellar surface

        # Scale from stellar surface to 1 AU
        fl_arr *= (R_star/AU)**2

        # Convert m-1 to nm-1
        fl_arr *= 1e-9

        # Convert W m-2 to erg s-1 cm-2
        fl_arr *= 1e3

    # Convert wavelengths to nm
    wl_arr *= 1e9

    # Return as lists
    return list(wl_arr), list(fl_arr)

def calc_star_luminosity(tmp:float, R_star:float):
    '''
    Calculate star's bolometric luminosity.

    Assumes that the star emits like a blackbody.
    '''

    # Evaluate stefan-boltzmann law at Teff
    lum = 0.0
    if tmp > PLANCK_MIN_TEMPERATURE:
        lum = 4 * np.pi * R_star * R_star * const_sigma * (tmp**4)

    return lum


def calc_instellation(tmp:float, R_star:float, sep:float):
    '''
    Calculate planet's instellation based on the star's luminosity.

    Parameters
    -----------
        tmp : float
            Star's effective temperature
        R_star : float
            Star's radius [m]
        sep : float
            Planet-star separation [m]

    Returns
    -----------
        S_0 : float
            Instellation [W m-2]
    '''

    # Get luminosity
    L_bol = calc_star_luminosity(tmp, R_star)

    # Return flux at planet-star distance
    return L_bol / (4 * np.pi * sep * sep)



# Generic orbital dynamics stuff
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np

from proteus.utils.constants import AU, const_G, secs_per_day

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

def update_separation(hf_row:dict, sma:float, ecc:float):
    '''
    Calculate time-averaged orbital separation on an elliptical path.

    Converts from AU to metres.
    https://physics.stackexchange.com/a/715749

    Parameters
    -------------
        hf_row: dict
            Current helpfile row
        sma: float
            Semimajor axis [AU]
        ecc: float
            Eccentricity
    '''

    hf_row["separation"] = sma * AU *  (1 + 0.5*ecc*ecc)

def update_period(hf_row:dict, sma:float):
    '''
    Calculate orbital period on an elliptical path.

    Assuming that M_volatiles << M_star + M_mantle + M_core.
    https://en.wikipedia.org/wiki/Elliptic_orbit#Orbital_period

    Parameters
    -------------
        hf_row: dict
            Current helpfile row
        sma: float
            Semimajor axis [AU]
    '''

    # Total mass of system, kg
    M_total = hf_row["M_star"] + hf_row["M_tot"]

    # Sanity check
    if M_total < 1e3:
       log.error("Unreasonable star+planet mass: %.5e kg"%M_total)

    # Standard gravitational parameter (planet mass + star mass)
    mu = const_G * M_total

    # Semimajor axis in SI units
    a = sma * AU

    # Orbital period [seconds]
    hf_row["period"] = 2 * np.pi * (a*a*a/mu)**0.5


def update_tides(hf_row:dict, config:Config):
    '''
    Update power density associated with tides.
    '''

    if config.orbit.module == 'dummy':
        log.info("    tidal power density: %.1e W kg-1"%(config.orbit.dummy.H_tide))

    # [call to some other tidal heating module goes here]

    else:
        # no interior module selected
        if config.interior.tidal_heat:
            log.warning("Tidal heating is enabled but no orbit module was selected")


def run_orbit(hf_row:dict, config:Config):
    '''
    Update parameters relating to orbital evolution and tides.
    '''

    log.info("Evolve orbit...")

    # Update orbital separation
    update_separation(hf_row, config.orbit.semimajoraxis, config.orbit.eccentricity)

    # Update orbital period
    update_period(hf_row, config.orbit.semimajoraxis)
    log.info("    period: %.1f days"%(hf_row["period"]/secs_per_day))

    # Update tidal heating
    update_tides(hf_row, config)



from __future__ import annotations

from attrs import define, field
from attrs.validators import ge, in_, le

from ._converters import none_if_none


@define
class Escape:
    """Escape parameters, model selection.

    Attributes
    ----------
    module: str | None
        Escape module to use. Choices: "none", "dummy", "zephyrus".
    zephyrus: Zephyrus
        Parameters for zephyrus module.
    dummy: EscapeDummy
        Parameters for dummy escape module.
    """
    module: str | None = field(
        validator=in_((None, 'dummy', 'zephyrus')), converter=none_if_none
        )

    zephyrus: Zephyrus
    dummy: EscapeDummy


@define
class Zephyrus:
    """Parameters for Zephyrus module.

    Attributes
    ----------
    Pxuv: float
        Pressure at which XUV radiation become opaque in the planetary atmosphere [bar]
    efficiency: float
        Escape efficiency factor
    tidal: bool
        Tidal contribution enabled
    """
    Pxuv: float = field(default=1e-2, validator=[ge(1e-10), le(1e1)])
    efficiency: float = field(default=1.0,validator=(ge(0.0), le(1.0)))
    tidal: bool = field(default=False)


@define
class EscapeDummy:
    """Dummy module.

    Attributes
    ----------
    rate: float
        Bulk unfractionated escape rate [kg s-1]
    """
    rate: float = field(validator=ge(0))



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import netCDF4 as nc
import numpy as np
from cmcrameri import cm
from matplotlib.ticker import LogLocator

from proteus.utils.plot import latex_float, sample_output

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)

def plot_emission(output_dir:str, times:list, plot_format="pdf",
                cumulative=False, xmax=2e4):

    log.info("Plot emission spectrum")

    norm = mpl.colors.LogNorm(vmin=max(times[0],1), vmax=times[-1])
    sm = plt.cm.ScalarMappable(cmap=cm.batlowK_r, norm=norm)
    sm.set_array([])

    scale = 1.7
    fig,ax = plt.subplots(1,1, figsize=(4*scale,3*scale))

    # atmosphere level to plot emission from
    atm_lvl = 0

    # plotting
    al = 0.9
    lw = 1.2
    ymax = 1e-20 # updated below
    ymin = 1e20  # updated below

    for i, t in enumerate( times ):

        label = latex_float(t)+" yr"
        color = sm.to_rgba(t)

        atm_file = os.path.join(output_dir, "data", "%d_atm.nc"%t)
        ds = nc.Dataset(atm_file)

        x_arr = []
        w_arr = []
        y_arr = np.array(ds["ba_U_LW"][atm_lvl,:]) + np.array(ds["ba_U_SW"][atm_lvl,:])

        # reversed?
        if ds["bandmin"][1] < ds["bandmin"][0]:
            bandmin = np.array(ds["bandmin"][::-1])
            bandmax = np.array(ds["bandmax"][::-1])
            y_arr = y_arr[::-1]
        else:
            bandmin = np.array(ds['bandmin'][:])
            bandmax = np.array(ds['bandmax'][:])

        y_arr = y_arr * 1000.0

        if cumulative:
            # bin edges
            x_arr = [bandmin[0]*1e9]
            for b in bandmax:
                x_arr.append(b*1e9)

            # copy array
            b_arr = np.zeros(len(y_arr))
            b_arr[:] = y_arr[:]

            # calulate cumulative emission
            y_arr = [0.0]
            for y in b_arr:
                y_arr.append(y_arr[-1]+y)

            x_arr = np.array(x_arr)
            y_arr = np.array(y_arr)

        else:
            x_arr = 1e9 * (bandmin + bandmax) / 2.0
            w_arr = 1e9 * (bandmax - bandmin)
            y_arr = y_arr / w_arr

        ds.close()

        # find maximum x value
        imax = np.argmin(np.abs(x_arr-xmax)) + 1
        imax = min(imax, len(x_arr))
        x_arr = x_arr[:imax]
        w_arr = w_arr[:imax]
        y_arr = y_arr[:imax]

        ax.step(x_arr, y_arr, label=label, color=color, alpha=al, lw=lw, where='mid')
        ymax = max(ymax, np.amax(y_arr))
        ymin = min(ymin, np.amin(y_arr))

    # Decorate
    ax.legend( fontsize=8, fancybox=True, framealpha=0.5 )
    ax.set_yscale("log")
    if cumulative:
        ax.set_ylabel(r"Cumulative emission [erg s$^{-1}$ cm$^{-2}$]")
    else:
        ax.set_ylabel(r"Flux density [erg s$^{-1}$ cm$^{-2}$ nm$^{-1}$]")
    ax.yaxis.set_major_locator(LogLocator(numticks=100))

    ax.set_xlabel("Wavelength [nm]")
    ax.set_xscale("log")

    ax.set_xlim(left=np.amin(x_arr), right=np.amax(x_arr))
    ax.set_ylim(bottom=ymin/2, top=ymax*2)

    plt.close()
    plt.ioff()

    fpath = os.path.join(output_dir, "plot_emission.%s"%plot_format)
    fig.savefig(fpath, dpi=200, bbox_inches='tight')

def plot_emission_entry(handler: Proteus):

    plot_times, _ = sample_output(handler, tmin=1000.0)
    print("Snapshots:", plot_times)


    # Plot fixed set from above
    plot_emission(
        handler.directories["output"],
        plot_times,
        plot_format=handler.config.params.out.plot_fmt,
   )


if __name__ == '__main__':
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_emission_entry(handler)



from __future__ import annotations

import glob
import logging
import os
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from cmcrameri import cm
from matplotlib.ticker import LogLocator, MultipleLocator
from mpl_toolkits.axes_grid1 import make_axes_locatable

from proteus.atmos_clim.common import read_ncdf_profile

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)


def plot_atmosphere_cbar(output_dir, plot_format="pdf", tmin=1e2):

    # Gather data files
    output_files = glob.glob(output_dir+"/data/*_atm.nc")
    output_times = [ int(str(f).split('/')[-1].split('_')[0]) for f in output_files]
    sort_mask = np.argsort(output_times)
    sorted_files = np.array(output_files)[sort_mask]
    sorted_times = np.array(output_times)[sort_mask]

    if np.amax(sorted_times) < tmin:
        tmin = 1

    if np.amax(sorted_times) < 2:
        log.debug("Insufficient data to make plot_atmosphere_cbar")
        return

    log.info("Plot atmosphere temperatures colourbar")

    # Parse NetCDF files
    stride = 1
    sorted_p = []
    sorted_t = []
    sorted_z = []
    sorted_y = []
    for i in range(0,len(sorted_files),stride):
        time = sorted_times[i]
        if time < tmin:
            continue
        sorted_y.append(sorted_times[i])

        prof = read_ncdf_profile(sorted_files[i])
        sorted_p.append(prof["p"] / 1.0e5)  # Convert Pa -> bar
        sorted_t.append(prof["t"])
        sorted_z.append(prof["z"] / 1.0e3)  # Convert m -> km
    nfiles = len(sorted_p)

    # Initialise plot
    scale = 1.1
    fig,ax = plt.subplots(1,1,figsize=(5*scale,4*scale))
    ax.set_ylabel("Pressure [bar]")
    ax.set_xlabel("Temperature [K]")
    ax.invert_yaxis()
    ax.set_yscale("log")

    # Colour mapping
    norm = mpl.colors.LogNorm(vmin=sorted_y[0], vmax=sorted_y[-1])
    sm = plt.cm.ScalarMappable(cmap=cm.batlowK_r, norm=norm)
    sm.set_array([])

    # Plot data
    for i in range(nfiles):
        a = 0.7
        c = sm.to_rgba(sorted_y[i])
        ax.plot(sorted_t[i], sorted_p[i], color=c, alpha=a, zorder=3)

    # Grid
    ax.grid(alpha=0.2, zorder=2)
    ax.set_xlim(0,np.amax(sorted_t)+100)
    ax.xaxis.set_minor_locator(MultipleLocator(base=250))

    ax.set_ylim(np.amax(sorted_p)*3, np.amin(sorted_p)/3)
    ax.yaxis.set_major_locator(LogLocator())

    # Plot colourbar
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    cbar = fig.colorbar(sm, cax=cax, orientation='vertical')
    cbar.set_label("Time [yr]")

    # Save plot
    fname = os.path.join(output_dir,"plot_atmosphere_cbar.%s"%plot_format)
    fig.savefig(fname, bbox_inches='tight', dpi=200)


def plot_atmosphere_cbar_entry(handler: Proteus):
    # Plot fixed set from above
    plot_atmosphere_cbar(
        output_dir=handler.directories["output"],
        plot_format=handler.config.params.out.plot_fmt)


if __name__ == "__main__":
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_atmosphere_cbar_entry(handler)



# Dummy atmosphere module
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np

from proteus.utils.constants import const_sigma
from proteus.utils.helper import UpdateStatusfile

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

# Run the dummy atmosphere module
def RunDummyAtm( dirs:dict, config:Config, T_magma:float, F_ins:float, R_int:float, M_int:float, P_surf:float):
    log.info("Running dummy atmosphere...")

    # Gamma factor: VERY simple parameterisation for the radiative properties of the atmosphere.
    # It represents a measure of the radiating temperature of the atmosphere above the
    #    surface, relative to the surface temperature itself
    # Setting this to 0 will result in an entirely transparent atmosphere
    # Setting this to 1 will result in an OLR of zero
    gamma           = 0.7

    # Parameters
    zenith_angle    = config.orbit.zenith_angle
    albedo_pl       = config.atmos_clim.albedo_pl
    inst_sf         = config.orbit.s0_factor
    albedo_s        = config.atmos_clim.surf_greyalbedo
    skin_d          = config.atmos_clim.surface_d
    skin_k          = config.atmos_clim.surface_k

    log.info("Gamma = %.4f" % gamma)
    if not (0.0 <= gamma <= 1.0):
        log.warning("Gamma value is out of bounds; expect unreasonable fluxes")

    # Simple rad trans
    def _calc_fluxes(x):
        # surface emission and stellar flux
        fl_U_LW = const_sigma * (x - gamma * x)**4.0
        fl_D_SW = F_ins * (1.0 - albedo_pl) * inst_sf * np.cos(zenith_angle * np.pi / 180.0)

        # surface reflection
        fl_U_SW = fl_D_SW * albedo_s
        fl_D_SW = fl_D_SW * (1.0-albedo_s)

        # net flux at surface
        fl_N = fl_U_LW + fl_U_SW - fl_D_SW

        return {"fl_U_LW":fl_U_LW, "fl_D_SW":fl_D_SW, "fl_U_SW":fl_U_SW, "fl_N":fl_N}

    # fixed T_Surf
    if config.atmos_clim.surf_state == 'fixed':
        log.info("Calculating fluxes with dummy atmosphere")
        T_surf_atm = T_magma
        fluxes = _calc_fluxes(T_surf_atm)

    # conductive lid
    elif config.atmos_clim.surf_state == 'skin':
        log.info("Calculating fluxes with dummy atmosphere and CBL")
        import scipy.optimize as optimise

        # We need to solve for the state where fl_N = f_skn
        # This function takes T_surf_atm as the input value, and returns fl_N - f_skn
        def _resid(x):
            F_skn = skin_k / skin_d * (T_magma - x)
            _f = _calc_fluxes(x)
            return _f["fl_N"] - F_skn

        r = optimise.root_scalar(_resid, method='secant', x0=T_magma, x1=T_magma-10.0,
                                        xtol=1.0e-7, maxiter=40)
        T_surf_atm = float(r.root)
        fluxes = _calc_fluxes(T_surf_atm)

        if r.converged:
            log.info("Found solution after %d iterations" % int(r.iterations))
        else:
            UpdateStatusfile(dirs, 22)
            raise Exception("Could not find solution for T_surf with dummy_atmosphere")

    else:
        UpdateStatusfile(dirs, 20)
        raise Exception("Invalid surface state chosen for dummy_atmosphere")

    # Require that the net flux must be upward
    F_atm_lim = fluxes["fl_N"]
    if config.atmos_clim.prevent_warming:
        F_atm_lim = max( 1.0e-8 , F_atm_lim )

    # Print if a limit was applied
    if not np.isclose(F_atm_lim , fluxes["fl_N"] ):
        log.warning("Change in F_atm [W m-2] limited in this step!")
        log.warning("    %g  ->  %g" % (fluxes["fl_N"] , F_atm_lim))

    # Return result
    log.info("Resultant values:")
    log.info("    T_surf =  %.3e  K"     % T_surf_atm)
    log.info("    F_atm  =  %.3e  W m-2" % F_atm_lim)
    log.info("    F_olr  =  %.3e  W m-2" % fluxes["fl_U_LW"])
    log.info("    F_sct  =  %.3e  W m-2" % fluxes["fl_U_SW"])

    # Escape level always at surface for dummy atmosphere
    if config.escape.module == 'zephyrus':
        log.warning("Setting escape level to surface because dummy atmosphere is used")

    output = {}
    output["T_surf"]  = T_surf_atm
    output["F_atm"]   = F_atm_lim             # Net flux at TOA
    output["F_olr"]   = fluxes["fl_U_LW"]     # OLR
    output["F_sct"]   = fluxes["fl_U_SW"]     # Scattered SW flux
    output["z_obs"]   = 0.0
    output["rho_obs"] = 3 * M_int / (4*np.pi*R_int**3)
    output["albedo"]  = fluxes["fl_U_SW"]/fluxes["fl_D_SW"]
    output["p_xuv"]   = P_surf
    output["z_xuv"]   = 0.0
    output["p_obs"]   = P_surf

    return output



# Function and classes used to run SPIDER
from __future__ import annotations

import glob
import json
import logging
import os
import platform
import subprocess
import sys
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from proteus.interior.timestep import next_step
from proteus.utils.constants import radnuc_data
from proteus.utils.helper import UpdateStatusfile, natural_sort, recursive_get

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)


class MyJSON( object ):

    '''load and access json data'''

    def __init__( self, filename ):
        self.filename = filename
        self._load()

    def _load( self ):
        '''load and store json data from file'''
        try:
            json_data  = open( self.filename )
        except FileNotFoundError:
            log.error('cannot find file: %s' % self.filename )
            log.error('please specify times for which data exists')
            sys.exit(1)
        self.data_d = json.load( json_data )
        json_data.close()

    # was get_field_data
    def get_dict( self, keys ):
        '''get all data relating to a particular field'''
        try:
            dict_d = recursive_get( self.data_d, keys )
            return dict_d
        except NameError:
            log.error('dictionary for %s does not exist', keys )
            sys.exit(1)

    # was get_field_units
    def get_dict_units( self, keys ):
        '''get the units (SI) of a particular field'''
        dict_d = recursive_get( self.data_d, keys )
        units = dict_d['units']
        units = None if units == 'None' else units
        return units

    # was get_scaled_field_values
    def get_dict_values( self, keys, fmt_o='' ):
        '''get the scaled values for a particular quantity'''
        dict_d = recursive_get( self.data_d, keys )
        scaling = float(dict_d['scaling'])
        if len( dict_d['values'] ) == 1:
            values_a = float( dict_d['values'][0] )
        else:
            values_a = np.array( [float(value) for value in dict_d['values']] )
        scaled_values_a = scaling * values_a
        if fmt_o:
            scaled_values_a = fmt_o.ascale( scaled_values_a )
        return scaled_values_a

    # was get_scaled_field_value_internal
    def get_dict_values_internal( self, keys, fmt_o='' ):
        '''get the scaled values for the internal nodes (ignore top
           and bottom nodes)'''
        scaled_values_a = self.get_dict_values( keys, fmt_o )
        return scaled_values_a[1:-1]

    def get_mixed_phase_boolean_array( self, nodes='basic' ):
        '''this array enables us to plot different linestyles for
           mixed phase versus single phase quantities'''
        if nodes == 'basic':
            phi = self.get_dict_values( ['data','phi_b'] )
        elif nodes == 'basic_internal':
            phi = self.get_dict_values_internal( ['data','phi_b'] )
        elif nodes == 'staggered':
            phi = self.get_dict_values( ['data','phi_s'] )
        # define mixed phase by these threshold values
        MIX = (phi<0.95) & (phi>0.05)
        return MIX

    def get_melt_phase_boolean_array( self, nodes='basic' ):
        '''this array enables us to plot different linestyles for
           melt phase regions'''
        if nodes == 'basic':
            phi = self.get_dict_values( ['data','phi_b'] )
        elif nodes == 'basic_internal':
            phi = self.get_dict_values_internal( ['data','phi_b'] )
        elif nodes == 'staggered':
            phi = self.get_dict_values( ['data','phi_s'] )
        MELT = (phi>0.95)
        return MELT

    def get_solid_phase_boolean_array( self, nodes='basic' ):
        '''this array enables us to plot different linestyles for
           solid phase regions'''
        if nodes == 'basic':
            phi = self.get_dict_values( ['data','phi_b'] )
        elif nodes == 'basic_internal':
            phi = self.get_dict_values_internal( ['data','phi_b'] )
        elif nodes == 'staggered':
            phi = self.get_dict_values( ['data','phi_s'] )
        SOLID = (phi<0.05)
        return SOLID

def read_jsons(output_dir:str, times:list):
    return [MyJSON(os.path.join(output_dir, "data", "%d.json"%t)) for t in times]

def get_all_output_times(odir:str):
    '''
    Get all times (in yr) from the json files located in the output directory
    '''

    odir = odir+'/data/'

    # locate times to process based on files located in odir/
    file_l = [f for f in os.listdir(odir) if os.path.isfile(odir+f)]
    if not file_l:
        raise Exception('Output data directory contains no files')

    time_l = [fname for fname in file_l]
    time_l = list(filter(lambda a: a.endswith('json'), time_l))
    time_l = [int(time.split('.json')[0]) for time in time_l]

    # ascending order
    time_l = sorted( time_l, key=int)
    time_a = np.array( time_l )

    return time_a

def get_dict_surface_values_for_specific_time( keys_t, time, indir='output'):
    '''Similar to above, but only loop over all times once and get
       all requested (surface / zero index) data in one go'''

    data_l = []

    filename = indir + '/data/{}.json'.format(time)
    myjson_o = MyJSON( filename )
    for key in keys_t:
        value = myjson_o.get_dict_values( key )
        data_l.append( value )


    return np.array(data_l)

#====================================================================
def _try_spider( dirs:dict, config:Config,
                IC_INTERIOR:int,
                hf_all:pd.DataFrame, hf_row:dict,
                step_sf:float, atol_sf:float ):
    '''
    Try to run spider with the current configuration.
    '''

    # Check that SPIDER can be found
    spider_exec = os.path.join(dirs["spider"],"spider")
    if not os.path.isfile(spider_exec):
        raise FileNotFoundError("SPIDER executable could not be found at '%s'"%spider_exec)

    # Bounds on tolereances
    step_sf = min(1.0, max(1.0e-10, step_sf))
    atol_sf = min(1.0e10, max(1.0e-10, atol_sf))

    # Recalculate time stepping
    if IC_INTERIOR == 2:

        # Current step number
        json_file   = MyJSON( dirs["output"]+'data/{}.json'.format(int(hf_row["Time"])) )
        step        = json_file.get_dict(['step'])

        # Get new time-step
        dtswitch = next_step(config, dirs, hf_row, hf_all, step_sf)

        # Number of total steps until currently desired switch/end time
        nsteps = 1
        nstepsmacro = step + nsteps
        dtmacro = dtswitch

        log.debug("Time options in RunSPIDER: dt=%.2e yrs in %d steps (at i=%d)" %
                                                    (dtmacro, nsteps, nstepsmacro))

    # For init loop
    else:
        nstepsmacro = 1
        dtmacro     = 0
        dtswitch    = 0

    empty_file = os.path.join(dirs["output"],"data", ".spider_tmp")
    open(empty_file, 'w').close()

    ### SPIDER base call sequence
    call_sequence = [
                        spider_exec,
                        "-options_file",           empty_file,
                        "-outputDirectory",        dirs["output"]+'data/',
                        "-IC_INTERIOR",            "%d"  %(IC_INTERIOR),
                        "-OXYGEN_FUGACITY_offset", "%.6e"%(config.outgas.fO2_shift_IW),  # Relative to the specified buffer
                        "-surface_bc_value",       "%.6e"%(hf_row["F_atm"]),
                        "-teqm",                   "%.6e"%(hf_row["T_eqm"]),
                        "-n",                      "%d"  %(config.interior.spider.num_levels),
                        "-nstepsmacro",            "%d"  %(nstepsmacro),
                        "-dtmacro",                "%.6e"%(dtmacro),
                        "-radius",                 "%.6e"%(hf_row["R_int"]),
                        "-gravity",                "%.6e"%(-1.0 * hf_row["gravity"]),
                        "-coresize",               "%.6e"%(config.struct.corefrac),
                        "-grain",                  "%.6e"%(config.interior.grain_size),
                    ]

    # Min of fractional and absolute Ts poststep change
    if hf_row["Time"] > 0:
        dTs_frac = config.interior.spider.tsurf_rtol * float(hf_all["T_surf"].iloc[-1])
        dT_int_max = np.min([ float(config.interior.spider.tsurf_atol), float(dTs_frac) ])
        call_sequence.extend(["-tsurf_poststep_change", str(dT_int_max)])
    else:
        call_sequence.extend(["-tsurf_poststep_change", str(config.interior.spider.tsurf_atol)])

    # set surface and core entropy (-1 is a flag to ignore)
    call_sequence.extend(["-ic_surface_entropy", "-1"])
    call_sequence.extend(["-ic_core_entropy",    "-1"])

    # Initial condition
    if IC_INTERIOR == 2:
        # get last JSON File
        last_filename = natural_sort([os.path.basename(x) for x in glob.glob(dirs["output"]+"data/*.json")])[-1]
        last_filename = os.path.join(dirs["output"], "data", last_filename)
        call_sequence.extend([
                                "-ic_interior_filename", str(last_filename),
                                "-activate_poststep",
                                "-activate_rollback"
                             ])
    else:
        # set to adiabat
        call_sequence.extend([
                                "-ic_adiabat_entropy", str(config.interior.spider.ini_entropy),
                                "-ic_dsdr", str(config.interior.spider.ini_dsdr) # initial dS/dr everywhere
                            ])

    # Mixing length parameterization: 1: variable | 2: constant
    call_sequence.extend(["-mixing_length", str(config.interior.spider.mixing_length)])
    call_sequence.extend(["-ts_sundials_atol", str(config.interior.spider.tolerance * atol_sf)])
    call_sequence.extend(["-ts_sundials_rtol", str(config.interior.spider.tolerance * atol_sf)])

    # Rollback
    call_sequence.extend(["-activate_poststep", "-activate_rollback"])

    # Dimensional scalings
    call_sequence.extend(["-radius0",   "63710000.0"])
    call_sequence.extend(["-entropy0",  "2993.025100070677"])
    call_sequence.extend(["-time0",     "1.0E5"])
    call_sequence.extend(["-pressure0", "10.0E5"])

    # Energy transport physics
    call_sequence.extend(["-CONDUCTION", "1"]) # conduction
    call_sequence.extend(["-CONVECTION", "1"]) # convection
    call_sequence.extend(["-MIXING    ", "1"]) # mixing (latent heat transport)
    call_sequence.extend(["-SEPARATION", "1"]) # gravitational separation of solid/melt

    # Tidal heating
    tidal_value = 0.0
    if config.interior.tidal_heat:
        if config.orbit.dummy:
            tidal_value = config.orbit.dummy.H_tide

        call_sequence.extend(["-HTIDAL",       "1"])
        call_sequence.extend(["-htidal_value", "%.5e"%tidal_value])

    # Properties lookup data (folder relative to SPIDER src)
    folder = "lookup_data/1TPa-dK09-elec-free/"
    call_sequence.extend(["-phase_names",  "melt,solid"])

    call_sequence.extend(["-melt_TYPE", "1"])
    call_sequence.extend(["-melt_alpha_filename_rel_to_src",          folder+"thermal_exp_melt.dat"])
    call_sequence.extend(["-melt_cp_filename_rel_to_src",             folder+"heat_capacity_melt.dat"])
    call_sequence.extend(["-melt_dTdPs_filename_rel_to_src",          folder+"adiabat_temp_grad_melt.dat"])
    call_sequence.extend(["-melt_rho_filename_rel_to_src",            folder+"density_melt.dat"])
    call_sequence.extend(["-melt_temp_filename_rel_to_src",           folder+"temperature_melt.dat"])
    call_sequence.extend(["-melt_phase_boundary_filename_rel_to_src", folder+"liquidus_A11_H13.dat"])
    call_sequence.extend(["-melt_log10visc", "2.0"])
    call_sequence.extend(["-melt_cond", "4.0"]) # conductivity of melt

    call_sequence.extend(["-solid_TYPE", "1"])
    call_sequence.extend(["-solid_alpha_filename_rel_to_src",            folder+"thermal_exp_solid.dat"])
    call_sequence.extend(["-solid_cp_filename_rel_to_src",               folder+"heat_capacity_solid.dat"])
    call_sequence.extend(["-solid_dTdPs_filename_rel_to_src",            folder+"adiabat_temp_grad_solid.dat"])
    call_sequence.extend(["-solid_rho_filename_rel_to_src",              folder+"density_solid.dat"])
    call_sequence.extend(["-solid_temp_filename_rel_to_src",             folder+"temperature_solid.dat"])
    call_sequence.extend(["-solid_phase_boundary_filename_rel_to_src",   folder+"solidus_A11_H13.dat"])
    call_sequence.extend(["-solid_log10visc", "22.0"])
    call_sequence.extend(["-solid_cond", "4.0"]) # conductivity of solid

    # static pressure profile derived from Adams-Williamson equation of state
    # these parameters are from fitting PREM in the lower mantle (for Earth)
    call_sequence.extend(["-adams_williamson_rhos", "4078.95095544"]) # surface density
    call_sequence.extend(["-adams_williamson_beta", "1.1115348931000002e-07"]) # beta parameter

    # eddy diffusivity
    # if negative, this value is adopted (units m^2/s)
    # if positive, this value is used to scale the internally calculated eddy diffusivity
    call_sequence.extend(["-eddy_diffusivity_thermal",  "1.0"])
    call_sequence.extend(["-eddy_diffusivity_chemical", "1.0"])


    # smoothing of material properties across liquidus and solidus
    # units of melt fraction (non-dimensional)
    call_sequence.extend(["-matprop_smooth_width", "1.0E-2"])

    # viscosity
    call_sequence.extend(["-phi_critical", "0.4"])  # transition melt fraction (non-dimensional)
    call_sequence.extend(["-phi_width",    "0.15"]) # transition width (non-dimensional)

    # core-mantle boundary condition
    call_sequence.extend(["-CORE_BC",  "1"])
    call_sequence.extend(["-rho_core", "%.6e"%(config.struct.core_density)]) # core density
    call_sequence.extend(["-cp_core",  "880.0"]) # core heat capacity

    # surface boundary condition
    # [4] heat flux (prescribe value using surface_bc_value)
    call_sequence.extend(["-SURFACE_BC", "4"])

    # parameterise the upper thermal boundary layer
    call_sequence.extend(["-PARAM_UTBL", "0"]) # disabled
    call_sequence.extend(["-param_utbl_const", "1.0E-7"]) # value of parameterisation

    # fO2 buffer chosen to define fO2 (7: Iron-Wustite)
    call_sequence.extend(["-OXYGEN_FUGACITY", "7"])

    # radionuclides
    if config.interior.radiogenic_heat:
        # offset by age_ini, which converts model simulation time to the actual age
        radio_t0 = config.delivery.radio_tref - config.star.age_ini
        radio_t0 *= 1e9 # Convert Gyr to yr
        radnuc_names = []

        def _append_radnuc(_iso, _cnc):
            radnuc_names.append(_iso)
            call_sequence.extend([f"-{_iso}_t0",              "%.5e"%radio_t0])
            call_sequence.extend([f"-{_iso}_concentration",   "%.5f"%_cnc])
            call_sequence.extend([f"-{_iso}_abundance",       "%.5e"%radnuc_data[_iso]["abundance"]])
            call_sequence.extend([f"-{_iso}_heat_production", "%.5e"%radnuc_data[_iso]["heatprod"]])
            call_sequence.extend([f"-{_iso}_half_life",       "%.5e"%radnuc_data[_iso]["halflife"]])

        if config.delivery.radio_K > 0.0:
            _append_radnuc("k40", config.delivery.radio_K)

        if config.delivery.radio_Th > 0.0:
            _append_radnuc("th232", config.delivery.radio_Th)

        if config.delivery.radio_U > 0.0:
            _append_radnuc("u235", config.delivery.radio_U)
            _append_radnuc("u238", config.delivery.radio_U)

        call_sequence.extend(["-radionuclide_names", ",".join(radnuc_names)])

    # Runtime info
    flags = ""
    for flag in call_sequence:
        flags += " " + flag
    # log.debug("SPIDER call sequence: '%s'" % flags)

    call_string = " ".join(call_sequence)

    # Environment
    spider_env = os.environ.copy()
    if platform.system() == "Darwin":
        spider_env["PETSC_ARCH"] = "arch-darwin-c-opt"
    else:
        spider_env["PETSC_ARCH"] = "arch-linux-c-opt"
    spider_env["PETSC_DIR"] = os.path.join(dirs["proteus"], "petsc")

    # Run SPIDER
    log.debug("SPIDER output suppressed")
    spider_print = open(dirs["output"]+"spider_recent.log",'w')
    spider_print.write(call_string+"\n")
    spider_print.flush()
    proc = subprocess.run([call_string],shell=True,stdout=spider_print, env=spider_env)
    spider_print.close()

    # Check status
    return bool(proc.returncode == 0)


def RunSPIDER( dirs:dict, config:Config, IC_INTERIOR:int,
              hf_all:pd.DataFrame, hf_row:dict ):
    '''
    Wrapper function for running SPIDER.
    This wrapper handles cases where SPIDER fails to find a solution.
    '''

    # parameters
    max_attempts = 5        # maximum number of attempts
    step_sf = 1.0           # step scale factor at attempt 1
    atol_sf = 1.0           # tolerance scale factor at attempt 1

    # tracking
    spider_success = False  # success?
    attempts = 0            # number of attempts so far

    # make attempts
    while not spider_success:
        attempts += 1
        log.debug("Attempt %d" % attempts)

        # run SPIDER
        spider_success = _try_spider(dirs, config, IC_INTERIOR, hf_all, hf_row, step_sf, atol_sf)

        if spider_success:
            # success
            log.debug("Attempt %d succeeded" % attempts)
        else:
            # failure
            log.warning("Attempt %d failed" % attempts)
            if attempts >= max_attempts:
                # give up
                log.error("Giving up")
                break
            else:
                # try again (change tolerance and step size)
                log.warning("Trying again")
                step_sf *= 0.5
                atol_sf *= 4.0

    # check status
    if spider_success:
        # success after some attempts
        return True
    else:
        # failure of all attempts
        UpdateStatusfile(dirs, 21)
        raise Exception("An error occurred when executing SPIDER (made %d attempts)" % attempts)


def ReadSPIDER(dirs:dict, config:Config, R_int:float):
    '''
    Read variables from last SPIDER output JSON file into a dictionary
    '''

    # Store variables in this dict
    output = {}

    ### Read in last SPIDER base parameters
    sim_time = get_all_output_times(dirs["output"])[-1]  # yr, as an integer value

    # SPIDER keys from JSON file that are read in
    keys_t = ( ('atmosphere','mass_liquid'),
                ('atmosphere','mass_solid'),
                ('atmosphere','mass_mantle'),
                ('atmosphere','mass_core'),
                ('atmosphere','temperature_surface'),
                ('rheological_front_phi','phi_global'),
                ('atmosphere','Fatm'),
                ('rheological_front_dynamic','depth'),
                )

    data_a = get_dict_surface_values_for_specific_time( keys_t, sim_time, indir=dirs["output"] )

    json_file = MyJSON( dirs["output"]+'/data/{}.json'.format(sim_time) )
    area_b   = json_file.get_dict_values(['data','area_b'])
    Hradio_s = json_file.get_dict_values(['data','Hradio_s'])
    Htidal_s = json_file.get_dict_values(['data','Htidal_s'])
    mass_s   = json_file.get_dict_values(['data','mass_s'])

    # Fill the new dict
    output["M_mantle_liquid"] = float(data_a[0])
    output["M_mantle_solid"]  = float(data_a[1])
    output["M_mantle"]        = float(data_a[2])

    # Surface properties
    output["T_magma"]         = float(data_a[4])
    output["Phi_global"]      = float(data_a[5])  # global melt fraction
    output["F_int"]           = float(data_a[6])  # Heat flux from interior
    output["RF_depth"]        = float(data_a[7])/R_int  # depth of rheological front

    # Tidal heating is not supported by SPIDER, so this should always be zero
    output["F_tidal"] = np.dot(Htidal_s, mass_s)/area_b[0]

    # Radiogenic heating
    output["F_radio"] = np.dot(Hradio_s, mass_s)/area_b[0]

    # Manually calculate heat flux at near-surface from energy gradient
    # Etot        = json_file.get_dict_values(['data','Etot_b'])
    # rad         = json_file.get_dict_values(['data','radius_b'])
    # area        = json_file.get_dict_values(['data','area_b'])
    # E0          = Etot[1] - (Etot[2]-Etot[1]) * (rad[2]-rad[1]) / (rad[1]-rad[0])
    # F_int2      = E0/area[0]

    # Limit F_int to positive values
    if config.atmos_clim.prevent_warming:
        output["F_int"] = max(1.0e-8, output["F_int"])

    # Check NaNs
    if np.isnan(output["T_magma"]):
        raise Exception("Magma ocean temperature is NaN")

    return sim_time, output



# Common atmosphere climate model functions
from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import netCDF4 as nc
import numpy as np

from proteus.utils.helper import find_nearest

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

def read_ncdf_profile(nc_fpath:str, extra_keys:list=[]):
    """Read data from atmosphere NetCDF output file.

    Automatically reads pressure (p), temperature (t), height (z) arrays with
    cell-centre (N) and cell-edge (N+1) values interleaved into a single combined array of
    length (2*N+1).

    Extra keys can be read-in using the extra_keys parameter. These will be stored with
    the same dimensions as in the NetCDF file.

    Parameters
    ----------
        nc_fpath : str
            Path to NetCDF file.

        extra_keys : list
            List of extra keys (strings) to read from the file.

    Returns
    ----------
        out : dict
            Dictionary containing numpy arrays of data from the file.
    """

    # open file
    ds = nc.Dataset(nc_fpath)

    p = np.array(ds.variables["p"][:])
    pl = np.array(ds.variables["pl"][:])

    t = np.array(ds.variables["tmp"][:])
    tl = np.array(ds.variables["tmpl"][:])

    z = np.array(ds.variables["z"][:])
    zl = np.array(ds.variables["zl"][:])

    nlev_c = len(p)

    # read pressure, temperature, height data into dictionary values
    out = {}
    out["p"] = [pl[0]]
    out["t"] = [tl[0]]
    out["z"] = [zl[0]]
    for i in range(nlev_c):
        out["p"].append(p[i])
        out["p"].append(pl[i+1])

        out["t"].append(t[i])
        out["t"].append(tl[i+1])

        out["z"].append(z[i])
        out["z"].append(zl[i+1])

    # Read extra keys
    for key in extra_keys:
        if key in ds.variables.keys():
            out[key] = np.array(ds.variables[key][:])

    # close file
    ds.close()

    # convert to np arrays
    for key in out.keys():
        out[key] = np.array(out[key], dtype=float)

    return out

def read_atmosphere_data(output_dir:str, times:list):
    """
    Read all p,t,z profiles from NetCDF files in a PROTEUS output folder.
    """
    profiles = [read_ncdf_profile(os.path.join(output_dir, "data", "%d_atm.nc"%t)) for t in times]
    if len(profiles) < 1:
        log.warning("No NetCDF files found in output folder")
    return profiles

def get_spfile_name_and_bands(config:Config):
    """
    Get spectral file name and bands from config
    """

    # Get table corresponding to the right atmosphere module
    obj = getattr(config.atmos_clim, config.atmos_clim.module)

    # Get bands and group name (strings)
    bands = obj.spectral_bands
    group = obj.spectral_group

    return group, bands


def get_spfile_path(fwl_dir:str, config:Config):
    """
    Get path to spectral file, given name and bands.
    """

    # Get group and bands (strings) from config
    group, bands = get_spfile_name_and_bands(config)

    # Construct file path
    return os.path.join(fwl_dir,"spectral_files",group,bands,group)+".sf"

def get_height_from_pressure(p_arr, z_arr, p_tgt):
    """
    Get the geometric height [m] corresponding to a given pressure.

    Parameters:
    ----------------
        p_arr: list
            Pressure array
        z_arr: list
            Height array
        p_tgt: float
            Target pressure

    Returns:
    ----------------
        p_close: float
            Closest pressure in the array
        z_close: float
            Closest height in the array
    """

    p_close, idx = find_nearest(p_arr, p_tgt)
    z_close = z_arr[idx]

    return float(p_close), z_close



"""This module describes the parameters for the data location, data output, and logging.
It also defines stopping criteria."""

from __future__ import annotations

from attrs import define, field
from attrs.validators import ge, gt, in_, lt


def max_bigger_than_min(instance, attribute, value):
    if value <= instance.minimum:
        raise ValueError("'maximum' has to be bigger than 'minimum'.")


@define
class Params:
    """Parameters for code execution, output files, time-stepping, convergence.

    Attributes
    ----------
    out: OutputParams
        Parameters for data / logging output.
    dt: TimeStepParams
        Parameters for time-stepping.
    stop: StopParams
        Parameters for stopping criteria.
    """
    out: OutputParams
    dt: TimeStepParams
    stop: StopParams

    resume: bool = field(default = False)
    offline: bool = field(default = False)


@define
class OutputParams:
    """Parameters for output files and logging

    Attributes
    ----------
    path: str
        Path to output folder relative to `PROTEUS/output/`.
    logging: str
        Log verbosity. Choices: 'INFO', 'DEBUG', 'ERROR', 'WARNING'.
    plot_mod: int
        Plotting frequency. 0: wait until completion. n: every n iterations.
    plot_fmt: str
        Plotting output file format. Choices: "png", "pdf".
    """
    path: str
    logging: str = field(validator=in_(('INFO', 'DEBUG', 'ERROR', 'WARNING')))
    plot_mod: int = field(validator=ge(0))
    plot_fmt: str = field(validator=in_(('pdf', 'png')))


@define
class TimeStepParams:
    """Parameters for time-stepping parameters

    Attributes
    ----------
    minimum: float
        Minimum time-step size [yr].
    maximum: float
        Maximum time-step size [yr].
    initial: float
        Initial time-step size [yr].
    starspec: float
        Maximum interval at which to recalculate the stellar spectrum [yr].
    starinst: float
        Maximum interval at which to recalculate instellation flux [yr].
    method: str
        Time-stepping method. Choices: 'proportional', 'adaptive', 'maximum'.
    proportional: DtProportional
        Parameters used to configure the proportional time-stepping scheme.
    adaptive: DtAdaptive
        Parameters used to configure the adaptive time-stepping scheme.
    """
    minimum: float = field(validator=gt(0))
    maximum: float = field(validator=gt(0))
    initial: float = field(validator=gt(0))
    starspec: float = field(validator=ge(0))
    starinst: float = field(validator=ge(0))
    method: str = field(validator=in_(('proportional', 'adaptive', 'maximum')))
    proportional: DtProportional
    adaptive: DtAdaptive


@define
class DtProportional:
    """Parameters used to configure the proportional time-stepping scheme.

    Attributes
    ----------
    propconst: float
        Proportionality constant.
    """
    propconst: float = field(validator=gt(0))


@define
class DtAdaptive:
    """Parameters used to configure the adaptive time-stepping scheme.

    Attributes
    ----------
    atol: float
        Absolute tolerance on time-step size [yr].
    rtol: float
        Relative tolerance on time-step size [dimensionless].
    """
    atol: float = field(validator=gt(0))
    rtol: float = field(validator=gt(0))


@define
class StopParams:
    """Parameters for termination criteria.

    Attributes
    ----------
    iters: StopIters
        Parameters for iteration number criteria.
    time: StopTime
        Parameters for maximum time criteria.
    solid: StopSolid
        Parameters for solidification criteria.
    radeqm: StopRadeqm
        Parameters for radiative equilibrium criteria.
    escape: StopEscape
        Parameters for escape criteria.
    """
    iters: StopIters
    time: StopTime
    solid: StopSolid
    radeqm: StopRadeqm
    escape: StopEscape


@define
class StopIters:
    """Parameters for iteration number criteria.

    Attributes
    ----------
    enabled: bool
        Enable criteria if True
    minimum: int
        Minimum number of iterations.
    maximum: int
        Maximum number of iterations.
    """
    enabled: bool
    minimum: int = field(validator=ge(0))
    maximum: int = field(validator=max_bigger_than_min)


@define
class StopTime:
    """Parameters for maximum time criteria.

    Attributes
    ----------
    enabled: bool
        Enable criteria if True
    minimum: float
        Model will absolutely not terminate until at least this time is reached [yr].
    maximum: float
        Model will terminate when this time is reached [yr].
    """
    enabled: bool
    minimum: float = field(validator=ge(0))
    maximum: float = field(validator=max_bigger_than_min)


@define
class StopSolid:
    """Parameters for solidification criteria.

    Attributes
    ----------
    enabled: bool
        Enable criteria if True.
    phi_crit: float
        Model will terminate when global melt fraction is less than this value [dimensionless].
    """
    enabled: bool
    phi_crit: float = field(validator=(gt(0), lt(1)))


@define
class StopRadeqm:
    """Parameters for radiative equilibrium stopping criteria.

    Attributes
    ----------
    enabled: bool
        Enable criteria if True
    atol: float
        Absolute tolerance on energy balance [W m-2].
    rtol: float
        Relative tolerance on energy balance.
    """
    enabled: bool
    atol: float = field(validator=gt(0))
    rtol: float = field(validator=ge(0))


@define
class StopEscape:
    """Parameters for escape stopping criteria.

    Attributes
    ----------
    enabled: bool
        Enable criteria if True
    mass_frac: float
        Model will termiante when atmosphere mass is less than this fraction of the initial atmosphere mass [dimensionless].
    """
    enabled: bool
    mass_frac: float = field(validator=(gt(0), lt(1)))
